import email.parser
import asyncio
from http.client import HTTPMessage, _MAXLINE, _MAXHEADERS
from io import BytesIO
from urllib.parse import parse_qs
from base64 import b64encode
from functools import reduce
from cgi import valid_boundary, parse_header

from pulsar import HttpException, BadRequest, isawaitable, ensure_future
from pulsar.utils.system import json
from pulsar.utils.structures import MultiValueDict, mapping_iterator
from pulsar.utils.httpurl import (DEFAULT_CHARSET, ENCODE_BODY_METHODS,
                                  JSON_CONTENT_TYPES, parse_options_header)


ONEMB = 2**20
# Default max size for body when not streaming
DEFAULT_MAXSIZE = 10*ONEMB

FORM_ENCODED_TYPES = ('application/x-www-form-urlencoded',
                      'application/x-url-encoded')
BODY_DATA = 0
BODY_FILES = 1
LARGE_BODY_CODE = 403


def http_protocol(parser):
    version = parser.get_version()
    return "HTTP/%s" % ".".join(('%s' % v for v in version))


class HttpBodyReader:
    _expect_sent = None
    _waiting = None

    def __init__(self, headers, parser, transport, **kw):
        self.headers = headers
        self.parser = parser
        self.reader = asyncio.StreamReader(**kw)
        self.reader.set_transport(transport)
        self.feed_data = self.reader.feed_data
        self.feed_eof = self.reader.feed_eof

    def waiting_expect(self):
        '''``True`` when the client is waiting for 100 Continue.
        '''
        if self._expect_sent is None:
            if (not self.reader.at_eof() and
                    self.headers.has('expect', '100-continue')):
                return True
            self._expect_sent = ''
        return False

    def can_continue(self):
        if self.waiting_expect():
            if self.parser.get_version() < (1, 1):
                raise HttpException(status=417)
            else:
                msg = '%s 100 Continue\r\n\r\n' % http_protocol(self.parser)
                self._expect_sent = msg
                self.reader._transport.write(msg.encode(DEFAULT_CHARSET))

    def fail(self):
        if self.waiting_expect():
            raise HttpException(status=417)

    def read(self, n=-1):
        self.can_continue()
        return self.reader.read(n=n)

    def readline(self):
        self.can_continue()
        return self.reader.readline()

    def readexactly(self, n):
        self.can_continue()
        return self.reader.readexactly(n)


def parse_form_data(environ, stream=None, **kw):
    '''Parse form data from an environ dict and return a (forms, files) tuple.

    Both tuple values are dictionaries with the form-field name as a key
    (unicode) and lists as values (multiple values per key are possible).
    The forms-dictionary contains form-field values as unicode strings.
    The files-dictionary contains :class:`MultipartPart` instances, either
    because the form-field was a file-upload or the value is to big to fit
    into memory limits.

    :parameter environ: A WSGI environment dict.
    :parameter stream: Optional callable accepting one parameter only, the
        instance of :class:`FormDecoder` being parsed. If provided, the
        callable is invoked when data or partial data has been successfully
        parsed.
    '''
    method = environ.get('REQUEST_METHOD', 'GET').upper()
    if method not in ENCODE_BODY_METHODS:
        raise HttpException(status=422)

    content_type = environ.get('CONTENT_TYPE')
    if content_type:
        content_type, options = parse_options_header(content_type)
        options.update(kw)
    else:
        options = kw

    if content_type == 'multipart/form-data':
        decoder = MultipartDecoder(environ, options, stream)
    elif content_type in FORM_ENCODED_TYPES:
        decoder = UrlEncodedDecoder(environ, options, stream)
    elif content_type in JSON_CONTENT_TYPES:
        decoder = JsonDecoder(environ, options, stream)
    else:
        decoder = BytesDecoder(environ, options, stream)

    return decoder.parse()


class FormDecoder:

    def __init__(self, environ, options, stream):
        self.environ = environ
        self.options = options
        self.stream = stream
        self.result = (MultiValueDict(), MultiValueDict())

    @property
    def content_length(self):
        return int(self.environ.get('CONTENT_LENGTH', -1))

    @property
    def charset(self):
        return self.options.get('charset', 'utf-8')

    def parse(self):
        raise NotImplementedError


class MultipartDecoder(FormDecoder):
    boundary = None

    def parse(self):
        boundary = self.options.get('boundary', '')
        if not valid_boundary(boundary):
            raise HttpException("Invalid boundary for multipart/form-data",
                                status=422)
        inp = self.environ.get('wsgi.input') or BytesIO()
        self.buffer = bytearray()

        if isinstance(inp, HttpBodyReader):
            return ensure_future(self._consume(inp, boundary),
                                 loop=inp.reader._loop)
        else:
            producer = BytesProducer(inp)
            return producer(self._consume, boundary)

    async def _consume(self, fp, boundary):
        sep = b'--'
        nextpart = ('--%s' % boundary).encode()
        lastpart = ("--%s--" % boundary).encode()
        terminator = b""

        while terminator != lastpart:
            nbytes = -1
            data = None
            current = None

            if terminator:
                headers = await parse_headers(fp)
                current = MultipartPart(self, headers)

                if nbytes > 0:
                    data = await fp.read(nbytes)
                else:
                    data = b''

                if current.name:
                    current.feed_data(data)
                else:
                    current = None

            while 1:
                line = await fp.readline()
                line = line or lastpart
                if line.startswith(sep):
                    terminator = line.rstrip()
                    if terminator in (nextpart, lastpart):
                        break
                if current:
                    current.feed_data(line)

            # Done with part.
            if current:
                current.done()

        self.environ['wsgi.input'] = BytesIO(self.buffer)
        return self.result


class BytesDecoder(FormDecoder):

    def parse(self, mem_limit=None, **kw):
        mem_limit = mem_limit or DEFAULT_MAXSIZE
        if self.content_length > mem_limit:
            raise HttpException("Request to big. Increase MAXMEM.",
                                status=LARGE_BODY_CODE)
        inp = self.environ.get('wsgi.input') or BytesIO()
        data = inp.read()

        if isawaitable(data):
            return ensure_future(self._async(data))
        else:
            return self._ready(data)

    async def _async(self, chunk):
        chunk = await chunk
        return self._ready(chunk)

    def _ready(self, data):
        self.environ['wsgi.input'] = BytesIO(data)
        self.result = (data, None)
        return self.result


class UrlEncodedDecoder(BytesDecoder):

    def _ready(self, data):
        self.environ['wsgi.input'] = BytesIO(data)
        charset = self.charset
        data = parse_qs(data.decode(charset), keep_blank_values=True)
        forms = self.result[0]
        for key, values in mapping_iterator(data):
            for value in values:
                forms[key] = value
        return self.result


class JsonDecoder(BytesDecoder):

    def _ready(self, data):
        self.environ['wsgi.input'] = BytesIO(data)
        try:
            self.result = (json.loads(data.decode(self.charset)), None)
        except Exception as exc:
            raise BadRequest('Could not decode JSON') from exc
        return self.result


class MultipartPart:
    filename = None
    name = ''

    def __init__(self, parser, headers):
        self.parser = parser
        self.headers = headers
        self._bytes = []
        self._done = False
        length = headers.get('content-length')
        content = headers.get('content-disposition')
        if length:
            try:
                length = int(length)
            except ValueError:
                length = -1
        else:
            length = -1
        self.length = length
        if content:
            key, params = parse_header(content)
            name = params.get('name')
            if key == 'form-data' and name:
                self.name = name
                self.filename = params.get('filename')

    def __repr__(self):
        return self.name

    @property
    def charset(self):
        return self.parser.charset

    @property
    def content_type(self):
        return self.headers.get('Content-Type')

    @property
    def size(self):
        return reduce(lambda x, y: x+len(y), self._bytes, 0)

    def bytes(self):
        '''Bytes'''
        return b''.join(self._bytes)

    def bytesio(self):
        return BytesIO(self.recv())

    def base64(self, charset=None):
        '''Data encoded as base 64'''
        return b64encode(self.bytes()).decode(charset or self.charset)

    def string(self, charset=None):
        '''Data decoded with the specified charset'''
        return self.bytes().decode(charset or self.charset)

    def complete(self):
        return self._done

    def feed_data(self, data):
        if data:
            self._bytes.append(data)
            if self.parser.stream:
                self.parser.stream(self)
            else:
                self.parser.buffer.extend(data)

    def recv(self, size=-1):
        if self._done:
            data = self._bytes
            self._bytes = []
        else:
            data = self._bytes[:-1]
            self._bytes = self._bytes[-1:]
        return b''.join(data)

    def is_file(self):
        return self.filename or self.content_type not in (None, 'text/plain')

    def done(self):
        if not self._done:

            if self._bytes and self.length < 0:
                # Strip final line terminator
                line = self._bytes[-1]
                if line[-2:] == b"\r\n":
                    line = line[:-2]
                elif line[-1:] == b"\n":
                    line = line[:-1]
                self._bytes[-1] = line

            self._done = True
            if self.parser.stream:
                self.parser.stream(self)

            if self.is_file():
                self.parser.result[1][self.name] = self
            else:
                self.parser.result[0][self.name] = self.string()


async def parse_headers(fp, _class=HTTPMessage):
    """Parses only RFC2822 headers from a file pointer.
    email Parser wants to see strings rather than bytes.
    But a TextIOWrapper around self.rfile would buffer too many bytes
    from the stream, bytes which we later need to read as bytes.
    So we read the correct bytes here, as bytes, for email Parser
    to parse.
    """
    headers = []
    while True:
        line = await fp.readline()
        if len(line) > _MAXLINE:
            raise HttpException("header line")
        headers.append(line)
        if len(headers) > _MAXHEADERS:
            raise HttpException("got more than %d headers" % _MAXHEADERS)
        if line in (b'\r\n', b'\n', b''):
            break
    hstring = b''.join(headers).decode('iso-8859-1')
    return email.parser.Parser(_class=_class).parsestr(hstring)


class BytesProducer:

    def __init__(self, bytes):
        self.bytes = bytes

    async def readline(self):
        return self.bytes.readline()

    async def read(self):
        return self.bytes.read()

    def __call__(self, consumer, *args):
        value = None
        coro = consumer(self, *args)
        while True:
            try:
                value = coro.send(value)
            except StopIteration as exc:
                value = exc.value
                break
        return value



import sys
import unittest

from pulsar import reraise


class TestMiscellaneous(unittest.TestCase):

    def test_reraise(self):
        self.assertRaises(RuntimeError, reraise, RuntimeError, RuntimeError())
        try:
            raise RuntimeError('bla')
        except Exception:
            exc_info = sys.exc_info()
        self.assertRaises(RuntimeError, reraise, *exc_info)



import asyncio
from inspect import isgenerator
from unittest import SkipTest, TestCase

from pulsar import ensure_future, isawaitable, HaltServer

from .utils import (TestFailure, skip_test, skip_reason,
                    expecting_failure, AsyncAssert, get_test_timeout)


class AbortTests(Exception):
    pass


class InvalidTestFunction(TestCase.failureException):
    pass


class Runner:

    def __init__(self, monitor, suite):
        cfg = suite.cfg
        loader = suite.loader
        tests = loader.test_files(cfg.labels, cfg.exclude_labels)
        #
        self._loop = monitor._loop
        self.cfg = suite.cfg
        self._time_start = self._loop.time()
        self.logger = monitor.logger
        self.suite = suite
        self.monitor = monitor
        self.loader = loader
        self.runner = loader.runner
        self.tests = list(reversed(tests))
        #
        self.runner.on_start()
        self._loop.call_soon(self._next_file)

    def _exit(self, exit_code):
        raise HaltServer(exit_code=exit_code)

    def _check_abort(self):
        if getattr(self._loop, 'exit_code', None):
            print('\nAbort tests')
            raise AbortTests

    def _next_file(self):
        if self.tests:
            tag, file_path_function = self.tests.pop()
            test_classes = iter(self.loader.import_module(*file_path_function))
            self._loop.call_soon(self._next_class, tag, test_classes)

        else:
            time_taken = self._loop.time() - self._time_start
            runner = self.runner
            runner.on_end()
            runner.printSummary(time_taken)
            if runner.result.errors or runner.result.failures:
                exit_code = 2
            else:
                exit_code = 0
            self._loop.call_soon(self._exit, exit_code)

    def _next_class(self, tag, test_classes):
        try:
            test_cls, test_fun = next(test_classes)
        except StopIteration:
            return self._loop.call_soon(self._next_file)

        test_cls.tag = tag
        test_cls.cfg = self.cfg
        test_cls.wait = AsyncAssert(test_cls)
        try:
            all_tests = self.runner.loadTestsFromTestCase(test_cls)
        except Exception:
            self.logger.exception('Could not load tests')
            all_tests = False
        else:
            all_tests = list(all_tests)
            if test_fun:
                all_tests, funcs = False, all_tests
                for fun in funcs:
                    if fun._testMethodName == test_fun:
                        all_tests = [fun]
                        break

        if all_tests:
            self.logger.info('Running %d Tests from %s.%s',
                             len(all_tests), tag, test_cls.__name__)
            self.runner.startTestClass(test_cls)
            coro = self._run_test_cls(test_cls, test_classes, all_tests)
            ensure_future(coro, loop=self._loop)
        else:
            self._loop.call_soon(self._next_class, tag, test_classes)

    async def _run_test_cls(self, test_cls, test_classes, all_tests):
        cfg = test_cls.cfg
        seq = getattr(test_cls, '_sequential_execution', cfg.sequential)
        test_timeout = get_test_timeout(test_cls, cfg.test_timeout)
        try:
            if skip_test(test_cls):
                raise SkipTest(skip_reason(test_cls))
            await self._run(test_cls.setUpClass, test_timeout)
        except SkipTest as exc:
            reason = str(exc)
            for test in all_tests:
                self.runner.addSkip(test, reason)
        except AbortTests:
            return
        except Exception as exc:
            self.logger.exception('Failure in setUpClass', exc_info=True)
            exc = TestFailure(exc)
            # setUpClass failed, fails all tests
            for test in all_tests:
                self.runner.startTest(test)
                self.add_failure(test, exc)
                self.runner.stopTest(test)
        else:
            try:
                if seq:
                    for test in all_tests:
                        await self._run_test(test, test_timeout)
                else:
                    await asyncio.wait([self._run_test(test, test_timeout)
                                        for test in all_tests],
                                       loop=self._loop)
            except AbortTests:
                return

        try:
            await self._run(test_cls.tearDownClass, test_timeout)
        except AbortTests:
            return
        except Exception:
            self.logger.exception('Failure in tearDownClass')

        self.logger.info('Finished tests from %s.%s',
                         test_cls.tag, test_cls.__name__)
        self._loop.call_soon(self._next_class, test_cls.tag, test_classes)

    async def _run(self, method, test_timeout):
        self._check_abort()
        coro = method()
        # a coroutine
        if isawaitable(coro):
            test_timeout = get_test_timeout(method, test_timeout)
            await asyncio.wait_for(coro, test_timeout, loop=self._loop)
        elif isgenerator(coro):
            raise InvalidTestFunction('test function returns a generator')

    async def _run_test(self, test, test_timeout):
        '''Run a ``test`` function using the following algorithm

        * Run :meth:`setUp` method in :attr:`testcls`
        * Run the test function
        * Run :meth:`tearDown` method in :attr:`testcls`
        '''
        error = None
        runner = self.runner
        runner.startTest(test)
        test_name = test._testMethodName
        method = getattr(test, test_name)
        if skip_test(method):
            reason = skip_reason(method)
            runner.addSkip(test, reason)
        else:
            error = await self._run_safe(test, 'setUp', test_timeout)
            if not error:
                test = runner.before_test_function_run(test)
                error = await self._run_safe(test, test_name, test_timeout)
                runner.after_test_function_run(test)
            error = await self._run_safe(test, 'tearDown', test_timeout, error)
            if not error:
                runner.addSuccess(test)
        runner.stopTest(test)

    async def _run_safe(self, test, method_name, test_timeout, error=None):
        self._check_abort()
        try:
            method = getattr(test, method_name)
            coro = method()
            # a coroutine
            if isawaitable(coro):
                test_timeout = get_test_timeout(method, test_timeout)
                await asyncio.wait_for(coro, test_timeout, loop=self._loop)
            elif isgenerator(coro):
                raise InvalidTestFunction('test function returns a generator')
        except SkipTest as exc:
            self.runner.addSkip(test, str(exc))
        except Exception as exc:
            if not error:
                error = TestFailure(exc)
                self.add_failure(test, error, expecting_failure(method))
        return error

    def add_failure(self, test, failure, expecting_failure=False):
        '''Add ``error`` to the list of errors.

        :param test: the test function object where the error occurs
        :param runner: the test runner
        :param error: the python exception for the error
        :param add_err: if ``True`` the error is added to the list of errors
        :return: a tuple containing the ``error`` and the ``exc_info``
        '''
        runner = self.runner
        if expecting_failure:
            runner.addExpectedFailure(test, failure)
        elif isinstance(failure.exc, test.failureException):
            runner.addFailure(test, failure)
        else:
            runner.addError(test, failure)



try:
    import signal
except ImportError:
    signal = None


__all__ = ['SIG_NAMES',
           'set_proctitle',
           'get_proctitle']


SIG_NAMES = {}

if signal:
    for sig in dir(signal):
        if sig.startswith('SIG') and sig[3] != "_":
            val = getattr(signal, sig)
            if isinstance(val, int):
                SIG_NAMES[val] = sig[3:]


try:
    from setproctitle import setproctitle, getproctitle

    def set_proctitle(title):
        setproctitle(title)
        return True

    def get_proctitle():
        return getproctitle()

except ImportError:  # pragma    nocover

    def set_proctitle(title):
        return False

    def get_proctitle():
        pass



import sys
import json
import logging
import asyncio
from collections import namedtuple

from pulsar import AsyncObject, as_coroutine, new_event_loop, ensure_future
from pulsar.utils.string import gen_unique_id
from pulsar.utils.tools import checkarity
from pulsar.apps.wsgi import Json
from pulsar.apps.http import HttpClient

from .handlers import RpcHandler, InvalidRequest, exception


__all__ = ['JSONRPC', 'JsonProxy', 'JsonBatchProxy']


logger = logging.getLogger('pulsar.jsonrpc')

BatchResponse = namedtuple('BatchResponse', 'id result exception')


class JSONRPC(RpcHandler):
    '''An :class:`.RpcHandler` for JSON-RPC services.

    Design to comply with the `JSON-RPC 2.0`_ Specification.

    JSON-RPC is a lightweight remote procedure call protocol
    designed to be simple.
    A remote method is invoked by sending a request to a remote service,
    the request is a single object serialised using JSON.

    .. _`JSON-RPC 2.0`: http://www.jsonrpc.org/specification
    '''
    version = '2.0'

    def __call__(self, request):
        return ensure_future(self._execute_request(request))

    async def _execute_request(self, request):
        response = request.response

        try:
            data = await as_coroutine(request.body_data())
        except ValueError:
            res, status = self._get_error_and_status(InvalidRequest(
                status=415, msg='Content-Type must be application/json'))
        else:
            # if it's batch request
            if isinstance(data, list):
                status = 200

                tasks = [self._call(request, each) for each in data]
                result = await asyncio.gather(*tasks)
                res = [r[0] for r in result]
            else:
                res, status = await self._call(request, data)

        response.status_code = status
        return Json(res).http_response(request)

    async def _call(self, request, data):
        exc_info = None
        proc = None
        try:
            if (not isinstance(data, dict) or
                    data.get('jsonrpc') != self.version or
                    'id' not in data):
                raise InvalidRequest(
                    'jsonrpc must be supplied and equal to "%s"' %
                    self.version
                )
            params = data.get('params')
            if isinstance(params, dict):
                args, kwargs = (), params
            else:
                args, kwargs = tuple(params or ()), {}
            #
            proc = self.get_handler(data.get('method'))
            result = await as_coroutine(proc(request, *args, **kwargs))
        except Exception as exc:
            result = exc
            exc_info = sys.exc_info()
        else:
            try:
                json.dumps(result)
            except Exception as exc:
                result = exc
                exc_info = sys.exc_info()
        #
        if exc_info:
            if isinstance(result, TypeError) and proc:
                msg = checkarity(proc, args, kwargs, discount=1)
            else:
                msg = None

            rpc_id = data.get('id') if isinstance(data, dict) else None

            res, status = self._get_error_and_status(
                result, msg=msg, rpc_id=rpc_id, exc_info=exc_info)
        else:
            res = {
                'id': data.get('id'),
                'jsonrpc': self.version,
                'result': result
            }
            status = 200

        return res, status

    def _get_error_and_status(self, exc, msg=None, rpc_id=None,
                              exc_info=None):
        res = {'id': rpc_id, 'jsonrpc': self.version}

        code = getattr(exc, 'fault_code', None)
        if not code:
            code = -32602 if msg else -32603
        msg = msg or str(exc) or 'JSON RPC exception'
        if code == -32603:
            logger.error(msg, exc_info=exc_info)
        else:
            logger.warning(msg)
        res['error'] = {
            'code': code,
            'message': msg,
            'data': getattr(exc, 'data', '')
        }

        return res, getattr(exc, 'status', 400)


class JsonCall:
    slots = ('_client', '_name')

    def __init__(self, client, name):
        self._client = client
        self._name = name

    def __repr__(self):
        return self._name
    __str__ = __repr__

    @property
    def url(self):
        return self._client.url

    @property
    def name(self):
        return self._name

    def __getattr__(self, name):
        name = "%s%s%s" % (self._name, self._client.separator, name)
        return self.__class__(self._client, name)

    def __call__(self, *args, **kwargs):
        result = self._client._call(self._name, *args, **kwargs)
        if self._client.sync:
            return self._client._loop.run_until_complete(result)
        else:
            return result


class JsonProxy(AsyncObject):
    '''A python Proxy class for :class:`.JSONRPC` Servers.

    :param url: server location
    :param version: JSON-RPC server version. Default ``2.0``
    :param id: optional request id, generated if not provided.
        Default ``None``.
    :param data: Extra data to include in all requests. Default ``None``.
    :param full_response: return the full Http response rather than
        just the content.
    :param http: optional http client. If provided it must have the ``request``
        method available which must be of the form::

            http.request(url, body=..., method=...)

        Default ``None``.
    :param encoding: encoding of the request. Default ``ascii``.

    Lets say your RPC server is running at ``http://domain.name.com/``::

        >>> a = JsonProxy('http://domain.name.com/')
        >>> a.add(3,4)
        7
        >>> a.ping()
        'pong'

    '''
    separator = '.'
    default_version = '2.0'
    default_timeout = 30

    def __init__(self, url, version=None, data=None,
                 full_response=False, http=None, timeout=None, sync=False,
                 loop=None, encoding='ascii', **kw):
        self.sync = sync
        self._url = url
        self._version = version or self.__class__.default_version
        self._full_response = full_response
        self._data = data if data is not None else {}
        if not http:
            timeout = timeout if timeout is not None else self.default_timeout
            if sync and not loop:
                loop = new_event_loop()
            http = HttpClient(timeout=timeout, loop=loop, **kw)
        http.headers['accept'] = 'application/json, text/*; q=0.5'
        http.headers['content-type'] = 'application/json'
        self._http = http
        self._encoding = encoding

    @property
    def url(self):
        return self._url

    @property
    def version(self):
        return self._version

    @property
    def _loop(self):
        return self._http._loop

    def makeid(self):
        '''Can be re-implemented by your own Proxy'''
        return gen_unique_id()

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__url)

    def __str__(self):
        return self.__repr__()

    def __getattr__(self, name):
        return JsonCall(self, name)

    async def _call(self, name, *args, **kwargs):
        data = self._get_data(name, *args, **kwargs)
        is_ascii = self._encoding == 'ascii'
        body = json.dumps(data, ensure_ascii=is_ascii).encode(self._encoding)
        resp = await self._http.post(self._url, data=body)
        if self._full_response:
            return resp
        else:
            content = resp.json()
            if resp.is_error:
                if 'error' not in content:
                    resp.raise_for_status()
            return self.loads(content)

    def _get_data(self, func_name, *args, **kwargs):
        id = self.makeid()
        params = self.get_params(*args, **kwargs)
        data = {'method': func_name, 'params': params, 'id': id,
                'jsonrpc': self._version}
        return data

    def get_params(self, *args, **kwargs):
        '''
        Create an array or positional or named parameters
        Mixing positional and named parameters in one
        call is not possible.
        '''
        kwargs.update(self._data)
        if args and kwargs:
            raise ValueError('Cannot mix positional and named parameters')
        if args:
            return list(args)
        else:
            return kwargs

    @staticmethod
    def loads(obj):
        if isinstance(obj, dict):
            if 'error' in obj:
                error = obj['error']
                raise exception(error.get('code'), error.get('message'))
            else:
                return obj.get('result')
        return obj


class JsonBatchProxy(JsonProxy):
    """A python Proxy class for :class:`.JSONRPC` Servers
    implementing batch protocol.

    :param url: server location
    :param version: JSON-RPC server version. Default ``2.0``
    :param id: optional request id, generated if not provided.
        Default ``None``.
    :param data: Extra data to include in all requests. Default ``None``.
    :param full_response: return the full Http response rather than
        just the response generator.
    :param http: optional http client. If provided it must have the ``request``
        method available which must be of the form::

            http.request(url, body=..., method=...)

        Default ``None``.
    :return: generator that returns batch response
        (a named tuple 'id result exception'). If ``full_response`` is True,
        then returns Http response.

    Lets say your RPC server is running at ``http://domain.name.com/``::

        >>> a = JsonBatchProxy('http://domain.name.com/')
        >>> a.add(3,4)
        'i86863002653c42278d7c5ff7506d84c7'
        >>> a.ping()
        'i71a9b79eef9b48eea2fb9d691c8e897e'

        >>> for each in (await a):
        >>>     print(each.id, each.result, each.exception)

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._batch = []

    def discard(self):
        """Clear pool of batch requests."""
        self._batch = []

    def __len__(self):
        return len(self._batch)

    def _call(self, name, *args, **kwargs):
        data = self._get_data(name, *args, **kwargs)
        is_ascii = self._encoding == 'ascii'
        body = json.dumps(data, ensure_ascii=is_ascii).encode(self._encoding)
        self._batch.append(body)
        return data['id']

    async def __call__(self):
        if not self._batch:
            return

        resp = await self._http.post(
            self._url, data=b'[' + b','.join(self._batch) + b']')

        if self._full_response:
            self.discard()
            return resp
        else:
            content = resp.json()
            if resp.is_error:
                if 'error' not in content:
                    resp.raise_for_status()
            self.discard()
            return self._response_gen(content)

    @staticmethod
    def _response_gen(content):
        if not isinstance(content, list):
            content = [content]

        for resp in content:
            try:
                yield BatchResponse(
                    id=resp['id'],
                    result=JsonBatchProxy.loads(resp),
                    exception=None
                )
            except Exception as err:
                yield BatchResponse(
                    id=resp['id'], result=None, exception=err
                )

    def __iter__(self):
        return self()



import unittest

from pulsar import Future
from pulsar.apps import wsgi
from pulsar.utils.system import json


class TestAsyncContent(unittest.TestCase):

    def test_string(self):
        a = wsgi.String('Hello')
        self.assertEqual(a.render(), 'Hello')
        self.assertRaises(RuntimeError, a.render)

    def test_simple_json(self):
        response = wsgi.Json({'bla': 'foo'})
        self.assertEqual(len(response.children), 1)
        self.assertEqual(response.content_type,
                         'application/json; charset=utf-8')
        self.assertFalse(response.as_list)
        self.assertEqual(response.render(), json.dumps({'bla': 'foo'}))

    def test_simple_json_as_list(self):
        response = wsgi.Json({'bla': 'foo'}, as_list=True)
        self.assertEqual(len(response.children), 1)
        self.assertEqual(response.content_type,
                         'application/json; charset=utf-8')
        self.assertTrue(response.as_list)
        self.assertEqual(response.render(), json.dumps([{'bla': 'foo'}]))

    def test_json_with_async_string(self):
        astr = wsgi.String('ciao')
        response = wsgi.Json({'bla': astr})
        self.assertEqual(len(response.children), 1)
        self.assertEqual(response.content_type,
                         'application/json; charset=utf-8')
        self.assertEqual(response.render(), json.dumps({'bla': 'ciao'}))

    async def test_json_with_async_string2(self):
        d = Future()
        astr = wsgi.String(d)
        response = wsgi.Json({'bla': astr})
        self.assertEqual(len(response.children), 1)
        result = response.render()
        self.assertIsInstance(result, Future)
        d.set_result('ciao')
        result = await result
        self.assertEqual(result, json.dumps({'bla': 'ciao'}))

    def test_append_self(self):
        root = wsgi.String()
        self.assertEqual(root.parent, None)
        root.append(root)
        self.assertEqual(root.parent, None)
        self.assertEqual(len(root.children), 0)

    def test_append(self):
        root = wsgi.String()
        child1 = wsgi.String()
        child2 = wsgi.String()
        root.append(child1)
        self.assertEqual(child1.parent, root)
        self.assertEqual(len(root.children), 1)
        root.prepend(child2)
        self.assertEqual(child2.parent, root)
        self.assertEqual(len(root.children), 2)

    def test_append_parent(self):
        root = wsgi.String()
        child1 = wsgi.String()
        child2 = wsgi.String()
        root.append(child1)
        root.append(child2)
        self.assertEqual(len(root.children), 2)
        child1.append(root)
        self.assertEqual(child1.parent, None)
        self.assertEqual(root.parent, child1)
        self.assertEqual(len(root.children), 1)
        self.assertEqual(len(child1.children), 1)

    def test_append_parent_with_parent(self):
        root = wsgi.String()
        child1 = wsgi.String()
        child2 = wsgi.String()
        child3 = wsgi.String()
        root.append(child1)
        child1.append(child2)
        child1.append(child3)
        self.assertEqual(len(root.children), 1)
        self.assertEqual(len(child1.children), 2)
        child2.append(child1)
        self.assertEqual(len(root.children), 1)
        self.assertEqual(root.children[0], child2)
        self.assertEqual(len(child2.children), 1)
        self.assertEqual(child1.parent, child2)
        self.assertEqual(child2.parent, root)

    def test_change_parent(self):
        root = wsgi.String()
        child1 = wsgi.String()
        child2 = wsgi.String()
        child3 = wsgi.String()
        root.append(child1)
        child1.append(child2)
        child1.append(child3)
        self.assertEqual(len(root.children), 1)
        self.assertEqual(len(child1.children), 2)
        root.append(child3)
        self.assertEqual(len(root.children), 2)
        self.assertEqual(len(child1.children), 1)

    def test_remove_valueerror(self):
        root = wsgi.String()
        child1 = wsgi.String()
        self.assertEqual(len(root.children), 0)
        root.remove(child1)
        self.assertEqual(len(root.children), 0)
        child1.append_to(root)
        self.assertEqual(len(root.children), 1)
        self.assertEqual(child1.parent, root)

    def test_remove_all(self):
        root = wsgi.Html('div')
        child1 = wsgi.Html('div')
        root.append(child1)
        root.append('ciao')
        self.assertEqual(len(root.children), 2)
        root.remove_all()
        self.assertEqual(len(root.children), 0)

    def test_media_path(self):
        media = wsgi.Scripts('/media/')
        self.assertTrue(media.is_relative('bla/test.js'))
        path = media.absolute_path('bla/foo.js')
        self.assertEqual(path, '/media/bla/foo.js')
        self.assertEqual(media.absolute_path('/bla/foo.js'), '/bla/foo.js')

    def test_links_minified(self):
        media = wsgi.Links('/media/', minified=True)
        self.assertEqual(media.absolute_path('bla/foo'),
                         '/media/bla/foo.min.css')
        self.assertEqual(media.absolute_path('bla/foo.min.css'),
                         '/media/bla/foo.min.css')
        self.assertEqual(media.absolute_path('bla/foo.css'),
                         '/media/bla/foo.css')

    def test_scripts_minified(self):
        media = wsgi.Scripts('/media/', minified=True)
        self.assertEqual(media.absolute_path('bla/foo'),
                         '/media/bla/foo.min.js')
        self.assertEqual(media.absolute_path('bla/foo.min.js'),
                         '/media/bla/foo.min.js')
        self.assertEqual(media.absolute_path('bla/foo.js'),
                         '/media/bla/foo.js')

    def test_html_doc_media(self):
        doc = wsgi.HtmlDocument(media_path='/foo/')
        self.assertEqual(doc.head.scripts.media_path, '/foo/')
        self.assertEqual(doc.head.links.media_path, '/foo/')
        doc.head.title = 'ciao'
        doc.head.media_path = '/assets/'
        self.assertEqual(doc.head.title, 'ciao')
        self.assertEqual(doc.head.scripts.media_path, '/assets/')
        self.assertEqual(doc.head.links.media_path, '/assets/')

    def test_link_condition(self):
        links = wsgi.Links('/media/')
        links.append('bla.css', condition='IE 6')
        html = links.render()
        lines = html.split('\n')
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], '<!--[if IE 6]>')
        self.assertEqual(lines[1], ("<link href='/media/bla.css' "
                                    "rel='stylesheet' type='text/css'>"))
        self.assertEqual(lines[2], '<![endif]-->')
        self.assertEqual(lines[3], '')



import os
import platform
import logging
import asyncio
from functools import partial
from collections import namedtuple
from io import StringIO, BytesIO
from urllib.parse import urlparse, parse_qsl, urlencode, urlunparse
from http.client import responses
try:
    import ssl
    BaseSSLError = ssl.SSLError
except ImportError:     # pragma    nocover
    ssl = None

    class BaseSSLError(Exception):
        pass

try:
    from certifi import where
    DEFAULT_CA_BUNDLE_PATH = where()
except ImportError:     # pragma    nocover
    DEFAULT_CA_BUNDLE_PATH = None

import pulsar
from pulsar import (AbortRequest, AbstractClient, Pool, Connection,
                    isawaitable, ProtocolConsumer, ensure_future,
                    HttpRequestException, HttpConnectionError, SSLError,
                    cfg_value)
from pulsar.utils import websocket
from pulsar.utils.system import json as _json
from pulsar.utils.pep import to_bytes
from pulsar.utils.structures import mapping_iterator
from pulsar.utils.httpurl import (http_parser, encode_multipart_formdata,
                                  Headers, get_environ_proxies, is_succesful,
                                  get_hostport, cookiejar_from_dict,
                                  host_no_default_port, http_chunks,
                                  parse_options_header,
                                  parse_header_links,
                                  JSON_CONTENT_TYPES)

from .plugins import (handle_cookies, WebSocket, Redirect,
                      Tunneling, TooManyRedirects, start_request,
                      response_content, keep_alive, HTTP11)

from .auth import Auth, HTTPBasicAuth, HTTPDigestAuth
from .oauth import OAuth1, OAuth2
from .stream import HttpStream, StreamConsumedError


__all__ = ['HttpRequest', 'HttpResponse', 'HttpClient', 'HTTPDigestAuth',
           'TooManyRedirects', 'Auth', 'OAuth1', 'OAuth2',
           'HttpStream', 'StreamConsumedError', 'full_url']


scheme_host = namedtuple('scheme_host', 'scheme netloc')
tls_schemes = ('https', 'wss')

LOGGER = logging.getLogger('pulsar.http')
FORM_URL_ENCODED = 'application/x-www-form-urlencoded'
MULTIPART_FORM_DATA = 'multipart/form-data'


def guess_filename(obj):
    """Tries to guess the filename of the given object."""
    name = getattr(obj, 'name', None)
    if name and name[0] != '<' and name[-1] != '>':
        return os.path.basename(name)


def scheme_host_port(url):
    url = urlparse(url)
    host, port = get_hostport(url.scheme, url.netloc)
    return url.scheme, host, port


def is_streamed(data):
    try:
        len(data)
    except TypeError:
        return True
    return False


def split_url_params(params):
    for key, values in mapping_iterator(params):
        if not isinstance(values, (list, tuple)):
            values = (values,)
        for value in values:
            yield key, value


def full_url(url, params, method=None):
    p = urlparse(url)
    if not p.netloc and method == 'CONNECT':
        p = urlparse('http://%s' % url)

    params = mapping_iterator(params)
    query = parse_qsl(p.query, True)
    query.extend(split_url_params(params))
    query = urlencode(query)
    return urlunparse((p.scheme, p.netloc, p.path,
                       p.params, query, p.fragment))


class RequestBase:
    inp_params = None
    release_connection = True
    history = None
    url = None
    scheme = None

    @property
    def unverifiable(self):
        """Unverifiable when a redirect.

        It is a redirect when :attr:`history` has past requests.
        """
        return bool(self.history)

    @property
    def address(self):
        return scheme_host_port(self.url)[1:]

    @property
    def origin_req_host(self):
        """Required by Cookies handlers
        """
        if self.history:
            return self.history[0].request.origin_req_host
        else:
            return scheme_host_port(self.url)[1]

    @property
    def type(self):
        return self.scheme

    @property
    def full_url(self):
        return self.url

    def get_full_url(self):
        """Required by Cookies handlers
        """
        return self.url

    def write_body(self, transport):
        pass


class HttpTunnel(RequestBase):
    first_line = None
    data = None

    def __init__(self, request, url):
        self.status = 0
        self.request = request
        self.url = '%s://%s:%s' % scheme_host_port(url)
        self.parser = request.parser
        request.new_parser()
        self.headers = request.client.tunnel_headers.copy()

    def __repr__(self):
        return 'Tunnel %s' % self.url
    __str__ = __repr__

    @property
    def key(self):
        return self.request.key

    @property
    def client(self):
        return self.request.client

    @property
    def version(self):
        return self.request.version

    def encode(self):
        req = self.request
        self.headers['host'] = req.get_header('host')
        bits = scheme_host_port(req.url)[1:] + (req.version,)
        self.first_line = 'CONNECT %s:%s %s\r\n' % bits
        return b''.join((self.first_line.encode('ascii'), bytes(self.headers)))

    def has_header(self, header_name):
        return header_name in self.headers

    def get_header(self, header_name, default=None):
        return self.headers.get(header_name, default)

    def remove_header(self, header_name):
        self.headers.pop(header_name, None)

    def apply(self, response, handler):
        """Tunnel the connection if needed
        """
        connection = response.connection
        tunnel_status = getattr(connection, '_tunnel_status', 0)
        if not tunnel_status:
            connection._tunnel_status = 1
            response.bind_event('pre_request', handler)
        elif tunnel_status == 1:
            connection._tunnel_status = 2
            response._request = self
            response.bind_event('on_headers', handler.on_headers)


class HttpRequest(RequestBase):
    """An :class:`HttpClient` request for an HTTP resource.

    This class has a similar interface to :class:`urllib.request.Request`.

    :param files: optional dictionary of name, file-like-objects.
    :param allow_redirects: allow the response to follow redirects.

    .. attribute:: method

        The request method

    .. attribute:: version

        HTTP version for this request, usually ``HTTP/1.1``

    .. attribute:: history

        List of past :class:`.HttpResponse` (collected during redirects).

    .. attribute:: wait_continue

        if ``True``, the :class:`HttpRequest` includes the
        ``Expect: 100-Continue`` header.

    .. attribute:: stream

        Allow for streaming body

    """
    _proxy = None
    _ssl = None
    _tunnel = None
    _write_done = False

    def __init__(self, client, url, method, inp_params=None, headers=None,
                 data=None, files=None, json=None, history=None, auth=None,
                 charset=None, max_redirects=10, source_address=None,
                 allow_redirects=False, decompress=True, version=None,
                 wait_continue=False, websocket_handler=None, cookies=None,
                 params=None, stream=False, proxies=None, verify=True,
                 **ignored):
        self.client = client
        self.method = method.upper()
        self.inp_params = inp_params or {}
        self.unredirected_headers = Headers(kind='client')
        self.history = history
        self.wait_continue = wait_continue
        self.max_redirects = max_redirects
        self.allow_redirects = allow_redirects
        self.charset = charset or 'utf-8'
        self.version = version
        self.decompress = decompress
        self.websocket_handler = websocket_handler
        self.source_address = source_address
        self.stream = stream
        self.verify = verify
        self.new_parser()
        if auth and not isinstance(auth, Auth):
            auth = HTTPBasicAuth(*auth)
        self.auth = auth
        self.headers = client._get_headers(headers)
        self.url = full_url(url, params, method=self.method)
        self.body = self._encode_body(data, files, json)
        self._set_proxy(proxies, ignored)
        cookies = cookiejar_from_dict(client.cookies, cookies)
        if cookies:
            cookies.add_cookie_header(self)

    @property
    def _loop(self):
        return self.client._loop

    @property
    def address(self):
        """``(host, port)`` tuple of the HTTP resource
        """
        return self._tunnel.address if self._tunnel else super().address

    @property
    def ssl(self):
        """Context for TLS connections.

        If this is a tunneled request and the tunnel connection is not yet
        established, it returns ``None``.
        """
        if not self._tunnel:
            return self._ssl

    @property
    def key(self):
        tunnel = self._tunnel.url if self._tunnel else None
        scheme, host, port = scheme_host_port(self._proxy or self.url)
        return scheme, host, port, tunnel, self.verify

    @property
    def proxy(self):
        """Proxy server for this request.
        """
        return self._proxy

    @property
    def tunnel(self):
        """Tunnel for this request.
        """
        return self._tunnel

    def __repr__(self):
        return self.first_line()
    __str__ = __repr__

    def first_line(self):
        p = urlparse(self.url)
        if self._proxy:
            if self.method == 'CONNECT':
                url = p.netloc
            else:
                url = self.url
        else:
            url = urlunparse(('', '', p.path or '/', p.params,
                              p.query, p.fragment))
        return '%s %s %s' % (self.method, url, self.version)

    def new_parser(self):
        self.parser = self.client.http_parser(kind=1,
                                              decompress=self.decompress,
                                              method=self.method)

    def is_chunked(self):
        return self.body and 'content-length' not in self.headers

    def encode(self):
        """The bytes representation of this :class:`HttpRequest`.

        Called by :class:`HttpResponse` when it needs to encode this
        :class:`HttpRequest` before sending it to the HTTP resource.
        """
        # Call body before fist_line in case the query is changes.
        first_line = self.first_line()
        if self.body and self.wait_continue:
            self.headers['expect'] = '100-continue'
        headers = self.headers
        if self.unredirected_headers:
            headers = self.unredirected_headers.copy()
            headers.update(self.headers)
        buffer = [first_line.encode('ascii'), b'\r\n',  bytes(headers)]
        return b''.join(buffer)

    def add_header(self, key, value):
        self.headers[key] = value

    def has_header(self, header_name):
        """Check ``header_name`` is in this request headers.
        """
        return (header_name in self.headers or
                header_name in self.unredirected_headers)

    def get_header(self, header_name, default=None):
        """Retrieve ``header_name`` from this request headers.
        """
        return self.headers.get(
            header_name, self.unredirected_headers.get(header_name, default))

    def remove_header(self, header_name):
        """Remove ``header_name`` from this request.
        """
        val1 = self.headers.pop(header_name, None)
        val2 = self.unredirected_headers.pop(header_name, None)
        return val1 or val2

    def add_unredirected_header(self, header_name, header_value):
        self.unredirected_headers[header_name] = header_value

    def write_body(self, transport):
        assert not self._write_done, 'Body already sent'
        self._write_done = True
        if not self.body:
            return
        if is_streamed(self.body):
            ensure_future(self._write_streamed_data(transport),
                          loop=self._loop)
        else:
            self._write_body_data(transport, self.body, True)

    # INTERNAL ENCODING METHODS
    def _encode_body(self, data, files, json):
        body = None
        if isinstance(data, (str, bytes)):
            if files:
                raise ValueError('data cannot be a string or bytes when '
                                 'files are present')
            body = to_bytes(data, self.charset)
        elif data and is_streamed(data):
            if files:
                raise ValueError('data cannot be an iterator when '
                                 'files are present')
            if 'content-length' not in self.headers:
                self.headers['transfer-encoding'] = 'chunked'
            return data
        elif data or files:
            if files:
                body, content_type = self._encode_files(data, files)
            else:
                body, content_type = self._encode_params(data)
            self.headers['Content-Type'] = content_type
        elif json:
            body = _json.dumps(json).encode(self.charset)
            self.headers['Content-Type'] = 'application/json'

        if body:
            self.headers['content-length'] = str(len(body))

        return body

    def _encode_files(self, data, files):
        fields = []
        for field, val in mapping_iterator(data or ()):
            if (isinstance(val, str) or isinstance(val, bytes) or
                    not hasattr(val, '__iter__')):
                val = [val]
            for v in val:
                if v is not None:
                    if not isinstance(v, bytes):
                        v = str(v)
                    fields.append((field.decode('utf-8') if
                                   isinstance(field, bytes) else field,
                                   v.encode('utf-8') if isinstance(v, str)
                                   else v))
        for (k, v) in mapping_iterator(files):
            # support for explicit filename
            ft = None
            if isinstance(v, (tuple, list)):
                if len(v) == 2:
                    fn, fp = v
                else:
                    fn, fp, ft = v
            else:
                fn = guess_filename(v) or k
                fp = v
            if isinstance(fp, bytes):
                fp = BytesIO(fp)
            elif isinstance(fp, str):
                fp = StringIO(fp)
            if ft:
                new_v = (fn, fp.read(), ft)
            else:
                new_v = (fn, fp.read())
            fields.append((k, new_v))
        #
        return encode_multipart_formdata(fields, charset=self.charset)

    def _encode_params(self, params):
        content_type = self.headers.get('content-type')
        # No content type given, chose one
        if not content_type:
            content_type = FORM_URL_ENCODED

        if hasattr(params, 'read'):
            params = params.read()

        if content_type in JSON_CONTENT_TYPES:
            body = _json.dumps(params)
        elif content_type == FORM_URL_ENCODED:
            body = urlencode(tuple(split_url_params(params)))
        elif content_type == MULTIPART_FORM_DATA:
            body, content_type = encode_multipart_formdata(
                params, charset=self.charset)
        else:
            body = params
        return to_bytes(body, self.charset), content_type

    def _write_body_data(self, transport, data, finish=False):
        if self.is_chunked():
            data = http_chunks(data, finish)
        elif data:
            data = (data,)
        else:
            return
        for chunk in data:
            transport.write(chunk)

    async def _write_streamed_data(self, transport):
        for data in self.body:
            if isawaitable(data):
                data = await data
            self._write_body_data(transport, data)
        self._write_body_data(transport, b'', True)

    # PROXY INTERNALS
    def _set_proxy(self, proxies, ignored):
        url = urlparse(self.url)
        self.unredirected_headers['host'] = host_no_default_port(url.scheme,
                                                                 url.netloc)
        if url.scheme in tls_schemes:
            self._ssl = self.client._ssl_context(verify=self.verify, **ignored)

        request_proxies = self.client.proxies.copy()
        if proxies:
            request_proxies.update(proxies)
        self.proxies = request_proxies
        #
        if url.scheme in request_proxies:
            host, port = get_hostport(url.scheme, url.netloc)
            no_proxy = [n for n in request_proxies.get('no', '').split(',')
                        if n]
            if not any(map(host.endswith, no_proxy)):
                url = request_proxies[url.scheme]
                if not self._ssl:
                    self._proxy = url
                else:
                    self._tunnel = HttpTunnel(self, url)


class HttpResponse(ProtocolConsumer):
    """A :class:`.ProtocolConsumer` for the HTTP client protocol.

    Initialised by a call to the :class:`HttpClient.request` method.

    There are two events you can yield in a coroutine:

    .. attribute:: on_headers

        fired once the response headers are received.

    .. attribute:: on_finished

        Fired once the whole request has finished

    Public API:
    """
    _tunnel_host = None
    _has_proxy = False
    _content = None
    _data_sent = None
    _status_code = None
    _cookies = None
    _raw = None
    request_again = None
    ONE_TIME_EVENTS = ('pre_request', 'on_headers', 'post_request')

    @property
    def parser(self):
        request = self.request
        if request:
            return request.parser

    def __repr__(self):
        return '<Response [%s]>' % (self.status_code or 'None')
    __str__ = __repr__

    @property
    def status_code(self):
        """Numeric status code such as 200, 404 and so forth.

        Available once the :attr:`on_headers` has fired.
        """
        return self._status_code

    @property
    def url(self):
        """The request full url.
        """
        request = self.request
        if request:
            return request.url

    @property
    def history(self):
        """List of :class:`.HttpResponse` objects from the history of the
        request. Any redirect responses will end up here.
        The list is sorted from the oldest to the most recent request."""
        request = self.request
        if request:
            return request.history

    @property
    def headers(self):
        if not hasattr(self, '_headers'):
            if self.parser and self.parser.is_headers_complete():
                self._headers = Headers(self.parser.get_headers())
        return getattr(self, '_headers', None)

    @property
    def is_error(self):
        if self.status_code:
            return not is_succesful(self.status_code)
        else:
            return False

    @property
    def ok(self):
        return not self.is_error

    @property
    def cookies(self):
        """Dictionary of cookies set by the server or ``None``.
        """
        return self._cookies

    @property
    def content(self):
        """Content of the response, in bytes
        """
        return response_content(self)

    @property
    def raw(self):
        """A raw asynchronous Http response
        """
        if self._raw is None:
            self._raw = HttpStream(self)
        return self._raw

    @property
    def links(self):
        """Returns the parsed header links of the response, if any
        """
        headers = self.headers or {}
        header = headers.get('link')
        l = {}
        if header:
            links = parse_header_links(header)
            for link in links:
                key = link.get('rel') or link.get('url')
                l[key] = link
        return l

    def recv_body(self):
        """Flush the response body and return it.
        """
        return self.parser.recv_body()

    def get_status(self):
        code = self.status_code
        if code:
            return '%d %s' % (code, responses.get(code, 'Unknown'))

    def text(self, charset=None, errors=None):
        """Decode content as a string.
        """
        data = self.content
        if data is not None:
            if charset is None:
                ct = self.headers.get('content-type')
                if ct:
                    ct, options = parse_options_header(ct)
                    charset = options.get('charset')
            return data.decode(charset or 'utf-8', errors or 'strict')
    content_string = text

    def json(self, charset=None):
        """Decode content as a JSON object.
        """
        return _json.loads(self.text(charset))

    def decode_content(self):
        """Return the best possible representation of the response body.
        """
        ct = self.headers.get('content-type')
        if ct:
            ct, options = parse_options_header(ct)
            charset = options.get('charset')
            if ct in JSON_CONTENT_TYPES:
                return self.json(charset)
            elif ct.startswith('text/'):
                return self.text(charset)
            elif ct == FORM_URL_ENCODED:
                return parse_qsl(self.content.decode(charset),
                                 keep_blank_values=True)
        return self.content

    def raise_for_status(self):
        """Raises stored :class:`HTTPError` or :class:`URLError`, if occurred.
        """
        if self.is_error:
            if self.status_code:
                raise HttpRequestException(response=self)
            else:
                raise HttpConnectionError(response=self,
                                          msg=self.on_finished.result.error)

    def info(self):
        """Required by python CookieJar.

        Return :attr:`headers`.
        """
        return self.headers

    # #####################################################################
    # #    PROTOCOL IMPLEMENTATION
    def start_request(self):
        request = self._request
        self.transport.write(request.encode())
        if request.headers.get('expect') != '100-continue':
            self.write_body()

    def data_received(self, data):
        request = self.request
        # request.parser my change (100-continue)
        # Always invoke it via request
        try:
            if request.parser.execute(data, len(data)) == len(data):
                if request.parser.is_headers_complete():
                    status_code = request.parser.get_status_code()
                    if (request.headers.has('expect', '100-continue') and
                            status_code == 100):
                        request.new_parser()
                        self.write_body()
                    else:
                        self._status_code = status_code
                        if not self.event('on_headers').fired():
                            self.fire_event('on_headers')
                        if (not self.event('post_request').fired() and
                                request.parser.is_message_complete()):
                            self.finished()
            else:
                raise pulsar.ProtocolError('%s\n%s' % (self, self.headers))
        except Exception as exc:
            self.finished(exc=exc)

    def write_body(self):
        self.request.write_body(self.transport)


class HttpClient(AbstractClient):
    """A client for HTTP/HTTPS servers.

    It handles pool of asynchronous connections.

    :param pool_size: set the :attr:`pool_size` attribute.
    :param store_cookies: set the :attr:`store_cookies` attribute

    .. attribute:: headers

        Default headers for this :class:`HttpClient`.

        Default: :attr:`DEFAULT_HTTP_HEADERS`.

    .. attribute:: cookies

        Default cookies for this :class:`HttpClient`.

    .. attribute:: store_cookies

        If ``True`` it remembers response cookies and sends them back to
        servers.

        Default: ``True``

    .. attribute:: timeout

        Default timeout for requests. If None or 0, no timeout on requests

    .. attribute:: proxies

        Dictionary of proxy servers for this client.

    .. attribute:: pool_size

        The size of a pool of connection for a given host.

    .. attribute:: connection_pools

        Dictionary of connection pools for different hosts

    .. attribute:: DEFAULT_HTTP_HEADERS

        Default headers for this :class:`HttpClient`

    """
    MANY_TIMES_EVENTS = ('connection_made', 'pre_request', 'on_headers',
                         'post_request', 'connection_lost')
    protocol_factory = partial(Connection, HttpResponse)
    max_redirects = 10
    """Maximum number of redirects.

    It can be overwritten on :meth:`request`.
    """
    connection_pool = Pool
    """Connection :class:`.Pool` factory
    """
    client_version = pulsar.SERVER_SOFTWARE
    """String for the ``User-Agent`` header.
    """
    version = HTTP11
    """Default HTTP request version for this :class:`HttpClient`.

    It can be overwritten on :meth:`request`.
    """
    DEFAULT_HTTP_HEADERS = Headers((
        ('Connection', 'Keep-Alive'),
        ('Accept', '*/*'),
        ('Accept-Encoding', 'deflate'),
        ('Accept-Encoding', 'gzip')),
        kind='client')
    DEFAULT_TUNNEL_HEADERS = Headers((
        ('Connection', 'Keep-Alive'),
        ('Proxy-Connection', 'Keep-Alive')),
        kind='client')
    request_parameters = ('max_redirects', 'decompress',
                          'websocket_handler', 'version',
                          'verify', 'stream')
    # Default hosts not affected by proxy settings. This can be overwritten
    # by specifying the "no" key in the proxies dictionary
    no_proxy = set(('localhost', platform.node()))

    def __init__(self, proxies=None, headers=None, verify=True,
                 cookies=None, store_cookies=True,
                 max_redirects=10, decompress=True, version=None,
                 websocket_handler=None, parser=None, trust_env=True,
                 loop=None, client_version=None, timeout=None, stream=False,
                 pool_size=10, frame_parser=None, logger=None,
                 close_connections=False, keep_alive=None):
        super().__init__(loop)
        self._logger = logger or LOGGER
        self.client_version = client_version or self.client_version
        self.connection_pools = {}
        self.pool_size = pool_size
        self.trust_env = trust_env
        self.timeout = timeout
        self.store_cookies = store_cookies
        self.max_redirects = max_redirects
        self.cookies = cookiejar_from_dict(cookies)
        self.decompress = decompress
        self.version = version or self.version
        self.verify = verify
        self.stream = stream
        self.close_connections = close_connections
        self.keep_alive = cfg_value('http_keep_alive', keep_alive)
        dheaders = self.DEFAULT_HTTP_HEADERS.copy()
        dheaders['user-agent'] = self.client_version
        if headers:
            dheaders.override(headers)
        self.headers = dheaders
        self.tunnel_headers = self.DEFAULT_TUNNEL_HEADERS.copy()
        self.proxies = dict(proxies or ())
        if not self.proxies and self.trust_env:
            self.proxies = get_environ_proxies()
            if 'no' not in self.proxies:
                self.proxies['no'] = ','.join(self.no_proxy)
        self.websocket_handler = websocket_handler
        self.http_parser = parser or http_parser
        self.frame_parser = frame_parser or websocket.frame_parser
        # Add hooks
        self.bind_event('pre_request', Tunneling(self._loop))
        self.bind_event('pre_request', WebSocket())
        self.bind_event('on_headers', handle_cookies)
        self.bind_event('post_request', Redirect())

    # API
    def connect(self, address):
        if isinstance(address, tuple):
            address = ':'.join(('%s' % v for v in address))
        return self.request('CONNECT', address)

    def get(self, url, **kwargs):
        """Sends a GET request and returns a :class:`HttpResponse` object.

        :params url: url for the new :class:`HttpRequest` object.
        :param \*\*kwargs: Optional arguments for the :meth:`request` method.
        """
        kwargs.setdefault('allow_redirects', True)
        return self.request('GET', url, **kwargs)

    def options(self, url, **kwargs):
        """Sends a OPTIONS request and returns a :class:`HttpResponse` object.

        :params url: url for the new :class:`HttpRequest` object.
        :param \*\*kwargs: Optional arguments for the :meth:`request` method.
        """
        kwargs.setdefault('allow_redirects', True)
        return self.request('OPTIONS', url, **kwargs)

    def head(self, url, **kwargs):
        """Sends a HEAD request and returns a :class:`HttpResponse` object.

        :params url: url for the new :class:`HttpRequest` object.
        :param \*\*kwargs: Optional arguments for the :meth:`request` method.
        """
        return self.request('HEAD', url, **kwargs)

    def post(self, url, **kwargs):
        """Sends a POST request and returns a :class:`HttpResponse` object.

        :params url: url for the new :class:`HttpRequest` object.
        :param \*\*kwargs: Optional arguments for the :meth:`request` method.
        """
        kwargs.setdefault('allow_redirects', True)
        return self.request('POST', url, **kwargs)

    def put(self, url, **kwargs):
        """Sends a PUT request and returns a :class:`HttpResponse` object.

        :params url: url for the new :class:`HttpRequest` object.
        :param \*\*kwargs: Optional arguments for the :meth:`request` method.
        """
        kwargs.setdefault('allow_redirects', True)
        return self.request('PUT', url, **kwargs)

    def patch(self, url, **kwargs):
        """Sends a PATCH request and returns a :class:`HttpResponse` object.

        :params url: url for the new :class:`HttpRequest` object.
        :param \*\*kwargs: Optional arguments for the :meth:`request` method.
        """
        kwargs.setdefault('allow_redirects', True)
        return self.request('PATCH', url, **kwargs)

    def delete(self, url, **kwargs):
        """Sends a DELETE request and returns a :class:`HttpResponse` object.

        :params url: url for the new :class:`HttpRequest` object.
        :param \*\*kwargs: Optional arguments for the :meth:`request` method.
        """
        kwargs.setdefault('allow_redirects', True)
        return self.request('DELETE', url, **kwargs)

    def request(self, method, url, timeout=None, **params):
        """Constructs and sends a request to a remote server.

        It returns a :class:`.Future` which results in a
        :class:`HttpResponse` object.

        :param method: request method for the :class:`HttpRequest`.
        :param url: URL for the :class:`HttpRequest`.
        :parameter response: optional pre-existing :class:`HttpResponse` which
            starts a new request (for redirects, digest authentication and
            so forth).
        :param params: optional parameters for the :class:`HttpRequest`
            initialisation.

        :rtype: a :class:`.Future`
        """
        response = self._request(method, url, **params)
        if timeout is None:
            timeout = self.timeout
        if timeout:
            response = asyncio.wait_for(response, timeout, loop=self._loop)
        if not self._loop.is_running():
            return self._loop.run_until_complete(response)
        else:
            return response

    def close(self):
        """Close all connections
        """
        waiters = []
        for p in self.connection_pools.values():
            waiters.append(p.close())
        self.connection_pools.clear()
        return asyncio.gather(*waiters, loop=self._loop)

    async def __aenter__(self):
        await self.close()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    # INTERNALS
    def create_protocol(self, **kw):
        kw['timeout'] = self.keep_alive
        return super().create_protocol(**kw)

    async def _request(self, method, url, **params):
        nparams = params.copy()
        nparams.update(((name, getattr(self, name)) for name in
                        self.request_parameters if name not in params))
        request = HttpRequest(self, url, method, params, **nparams)
        pool = self.connection_pools.get(request.key)
        if pool is None:
            host, port = request.address
            connector = partial(self.create_connection,
                                (host, port),
                                ssl=request.ssl)
            pool = self.connection_pool(connector, pool_size=self.pool_size,
                                        loop=self._loop)
            self.connection_pools[request.key] = pool
        try:
            conn = await pool.connect()
        except BaseSSLError as e:
            raise SSLError(str(e), response=self) from None
        except ConnectionRefusedError as e:
            raise HttpConnectionError(str(e), response=self) from None

        with conn:
            try:
                response = await start_request(request, conn)
                headers = response.headers
            except AbortRequest:
                response = None
                headers = None

            if (not headers or
                    not keep_alive(response.request.version, headers) or
                    response.status_code == 101 or
                    response.request.stream or
                    self.close_connections):
                conn.detach()

        # Handle a possible redirect
        if response and isinstance(response.request_again, tuple):
            method, url, params = response.request_again
            response = await self._request(method, url, **params)
        return response

    def _get_headers(self, headers=None):
        # Returns a :class:`Header` obtained from combining
        # :attr:`headers` with *headers*. Can handle websocket requests.
        d = self.headers.copy()
        if headers:
            d.override(headers)
        return d

    def _ssl_context(self, verify=True, cert_reqs=None,
                     check_hostname=False, certfile=None, keyfile=None,
                     cafile=None, capath=None, cadata=None, **kw):
        assert ssl, 'SSL not supported'
        cafile = cafile or DEFAULT_CA_BUNDLE_PATH

        if verify is True:
            cert_reqs = ssl.CERT_REQUIRED
            check_hostname = True

        if isinstance(verify, str):
            cert_reqs = ssl.CERT_REQUIRED
            if os.path.isfile(verify):
                cafile = verify
            elif os.path.isdir(verify):
                capath = verify

        return ssl._create_unverified_context(cert_reqs=cert_reqs,
                                              check_hostname=check_hostname,
                                              certfile=certfile,
                                              keyfile=keyfile,
                                              cafile=cafile,
                                              capath=capath,
                                              cadata=cadata)



import uuid
from asyncio import sleep

from pulsar import isawaitable, LockError, LockBase


class RedisScript:
    '''An executable Lua script object
    '''
    def __init__(self, script):
        self.script = script
        self.sha = None

    async def __call__(self, client, keys=None, args=None):
        '''Execute the script, passing any required ``args``
        '''
        if self.sha not in client.store.loaded_scripts:
            sha = await client.immediate_execute('SCRIPT', 'LOAD', self.script)
            self.sha = sha.decode('utf-8')
            client.store.loaded_scripts.add(self.sha)

        result = client.evalsha(self.sha, keys, args)
        if isawaitable(result):
            result = await result
        return result


class Lock(LockBase):
    """Asynchronous locking primitive for distributing computing
    """
    def __init__(self, client, name, timeout=None, blocking=True, sleep=0.2):
        super().__init__(name, loop=client._loop, timeout=timeout,
                         blocking=blocking)
        self._token = None
        self.client = client
        self.sleep = sleep
        if self.blocking:
            self.sleep = min(self.sleep, self.blocking)

    def locked(self):
        ''''Return the token that acquire the lock or None.
        '''
        return bool(self._token)

    async def _acquire(self):
        token = uuid.uuid1().hex.encode('utf-8')
        timeout = self.timeout and int(self.timeout * 1000) or ''
        acquired = await self.lua_acquire(self.client, keys=[self.name],
                                          args=[token, timeout])
        if acquired:
            self._token = token

        return acquired

    async def acquire(self):
        loop = self._loop
        start = loop.time()
        acquired = await self._acquire()
        if self.blocking is False:
            return acquired
        timeout = self.blocking
        if timeout is True:
            timeout = 0
        while not acquired:
            if timeout and loop.time() - start >= timeout:
                break
            await sleep(self.sleep)
            acquired = await self._acquire()

        return acquired

    async def release(self):
        expected_token = self._token
        if not expected_token:
            raise LockError("Cannot release an unlocked lock")
        released = await self.lua_release(self.client, keys=[self.name],
                                          args=[expected_token])
        self._token = None
        if not released:
            raise LockError("Cannot release a lock that's no longer owned")
        return True

    lua_acquire = RedisScript("""
        if redis.call('setnx', KEYS[1], ARGV[1]) == 1 then
            if ARGV[2] ~= '' then
                redis.call('pexpire', KEYS[1], ARGV[2])
            end
            return 1
        end
        return 0
    """)

    # KEYS[1] - lock name
    # ARGS[1] - token
    # return 1 if the lock was released, otherwise 0
    lua_release = RedisScript("""
        local token = redis.call('get', KEYS[1])
        if not token or token ~= ARGV[1] then
            return 0
        end
        redis.call('del', KEYS[1])
        return 1
    """)



'''Tests the "flaskapp" example.'''
import unittest

from pulsar import send, SERVER_SOFTWARE
from pulsar.apps.http import HttpClient

try:
    from examples.flaskapp.manage import server
except ImportError:
    server = None


@unittest.skipUnless(server, "Requires flask module")
class TestFlaskApp(unittest.TestCase):
    app_cfg = None

    @classmethod
    def name(cls):
        return 'flaskapptest'

    @classmethod
    async def setUpClass(cls):
        s = server(name=cls.name(), bind='127.0.0.1:0')
        cls.app_cfg = await send('arbiter', 'run', s)
        cls.uri = 'http://{0}:{1}'.format(*cls.app_cfg.addresses[0])
        cls.client = HttpClient()

    @classmethod
    def tearDownClass(cls):
        if cls.app_cfg is not None:
            return send('arbiter', 'kill_actor', cls.app_cfg.name)

    async def test_response_200(self):
        c = self.client
        response = await c.get(self.uri)
        self.assertEqual(response.status_code, 200)
        content = response.content
        self.assertEqual(content, b'Flask Example')
        headers = response.headers
        self.assertTrue(headers)
        self.assertEqual(headers['server'], SERVER_SOFTWARE)

    async def test_response_404(self):
        c = self.client
        response = await c.get('%s/bh' % self.uri)
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b'404 Page')



import unittest

from pulsar import get_actor

from tests.http import base


class ExternalBase(base.TestHttpClientBase):
    with_httpbin = False

    def after_response(self, response):
        pass

    async def test_get_https(self):
        client = self.client()
        response = await client.get('https://github.com/trending')
        self.assertEqual(response.status_code, 200)

    async def test_header_links_and_close(self):
        client = self.client()
        print(client._loop)
        baseurl = 'https://api.github.com/gists/public'
        response = await client.get(baseurl)
        if response.status_code == 403:
            # TODO: this fails in travis for some reason
            return
        self.assertEqual(response.status_code, 200)
        links = response.links
        self.assertTrue(links)
        next = links['next']
        self.assertTrue('rel' in next)
        self.assertTrue('url' in next)
        await client.close()


class ProxyExternal(ExternalBase):

    def after_response(self, response):
        self.assertTrue(response.request.proxy)

    async def test_get_https(self):
        client = self.client()
        response = await client.get('https://github.com/trending')
        self.assertEqual(response.status_code, 200)


@unittest.skipUnless(get_actor().cfg.http_proxy == '',
                     'Requires no external proxy')
class Test_HttpClient_NoProxy_External(ExternalBase, unittest.TestCase):
    '''Test external URI when no global proxy server is present.
    '''


@unittest.skipUnless(get_actor().cfg.http_proxy == '',
                     'Requires no external proxy')
class Test_HttpClient_Proxy_External(ProxyExternal, unittest.TestCase):
    with_proxy = True


@unittest.skipUnless(get_actor().cfg.http_proxy, 'Requires external proxy')
class Test_HttpClient_ExternalProxy_External(ProxyExternal,
                                             unittest.TestCase):
    pass



import binascii
import time
import unittest
import asyncio
import datetime

import pulsar
from pulsar.utils.string import random_string
from pulsar.utils.structures import Zset
from pulsar.apps.ds import PulsarDS, redis_parser, ResponseError
from pulsar.apps.data import create_store


class Listener:

    def __init__(self):
        self._messages = asyncio.Queue()

    def __call__(self, channel, message):
        self._messages.put_nowait((channel, message))

    def get(self):
        return self._messages.get()


class StringProtocol:

    def encode(self, message):
        return message

    def decode(self, message):
        return message.decode('utf-8')


class StoreMixin:
    redis_py_parser = False

    @classmethod
    def create_store(cls, address, namespace=None, pool_size=2, **kw):
        if cls.redis_py_parser:
            kw['parser_class'] = redis_parser(True)
        if not namespace:
            namespace = cls.randomkey(6).lower()
        return create_store(address, namespace=namespace,
                            pool_size=pool_size, **kw)

    @classmethod
    def randomkey(cls, length=None):
        return random_string(min_length=length, max_length=length)

    async def _remove_and_push(self, key, rem=1):
        c = self.client
        self.assertEqual(await c.delete(key), rem)
        self.assertEqual(await c.rpush(key, 'bla'), 1)

    async def _remove_and_sadd(self, key, rem=1):
        c = self.client
        self.assertEqual(await c.delete(key), rem)
        self.assertEqual(await c.sadd(key, 'bla'), 1)


class RedisCommands(StoreMixin):

    def test_store(self):
        store = self.store
        self.assertTrue(store.namespace)

    ###########################################################################
    #    KEYS
    async def test_dump_restore(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.dump(key), None)
        eq(await c.set(key, 'hello'), True)
        value = await c.dump(key)
        self.assertTrue(value)
        await self.wait.assertRaises(
            ResponseError, c.restore, key, 0, 'bla')
        eq(await c.restore(key+'2', 0, value), True)
        eq(await c.get(key+'2'), b'hello')

    async def test_exists(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.exists(key), False)
        eq(await c.set(key, 'hello'), True)
        eq(await c.exists(key), True)
        eq(await c.delete(key), 1)

    async def test_expire_persist_ttl(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        await self.wait.assertRaises(ResponseError, c.expire, key, 'bla')
        eq(await c.expire(key, 1), False)
        eq(await c.set(key, 1), True)
        eq(await c.expire(key, 10), True)
        ttl = await c.ttl(key)
        self.assertTrue(ttl > 0 and ttl <= 10)
        eq(await c.persist(key), True)
        eq(await c.ttl(key), -1)
        eq(await c.persist(key), False)

    async def test_expireat(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        await self.wait.assertRaises(ResponseError, c.expireat, key, 'bla')
        t = int(time.time() + 3)
        eq(await c.expireat(key, t), False)
        eq(await c.set(key, 1), True)
        t = int(time.time() + 10)
        eq(await c.expireat(key, t), True)
        ttl = await c.ttl(key)
        self.assertTrue(ttl > 0 and ttl <= 10)
        eq(await c.persist(key), True)
        eq(await c.ttl(key), -1)
        eq(await c.persist(key), False)

    async def test_keys(self):
        key = self.randomkey()
        keya = '%s_a' % key
        keyb = '%s_a' % key
        keyc = '%sc' % key
        c = self.client
        eq = self.assertEqual
        keys_with_underscores = set([keya.encode('utf-8'),
                                     keyb.encode('utf-8')])
        keys = keys_with_underscores.union(set([keyc.encode('utf-8')]))
        eq(await c.mset(keya, 1, keyb, 2, keyc, 3), True)
        k1 = await c.keys('%s_*' % key)
        k2 = await c.keys('%s*' % key)
        self.assertEqual(set(k1), keys_with_underscores)
        self.assertEqual(set(k2), keys)

    async def test_move(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        db = 3 if c.store.database == 4 else 4
        await self.wait.assertRaises(ResponseError, c.move, key, 'bla')
        eq(await c.move(key, db), False)
        eq(await c.set(key, 'ciao'), True)
        eq(await c.move(key, db), True)
        s2 = self.create_store(self.store.dns, database=db)
        c2 = s2.client()
        eq(await c2.get(key), b'ciao')
        eq(await c.exists(key), False)
        eq(await c.set(key, 'foo'), True)
        eq(await c.move(key, db), False)
        eq(await c.exists(key), True)

    async def __test_randomkey(self):
        # TODO: this test fails sometimes
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.mset(key, 1, key+'a', 2, key+'b', 3), True)
        key = await c.randomkey()
        eq(await c.exists(key), True)

    async def test_rename_renamenx(self):
        key = self.randomkey()
        des = self.randomkey()
        c = self.client
        eq = self.assertEqual
        await self.wait.assertRaises(ResponseError, c.rename, key, des)
        eq(await c.set(key, 'hello'), True)
        eq(await c.rename(key, des), True)
        eq(await c.exists(key), False)
        eq(await c.get(des), b'hello')
        eq(await c.set(key, 'ciao'), True)
        eq(await c.renamenx(key, des), False)
        eq(await c.renamenx(key, des+'a'), True)
        eq(await c.exists(key), False)

    ###########################################################################
    #    BAD REQUESTS
    # async def test_no_command(self):
    #     await self.wait.assertRaises(ResponseError, self.store.execute)

    # async def test_bad_command(self):
    #     await self.wait.assertRaises(ResponseError, self.store.execute,
    #                                   'foo')
    ###########################################################################
    #    STRINGS
    async def test_append(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.append(key, 'a1'), 2)
        eq(await c.get(key), b'a1')
        eq(await c.append(key, 'a2'), 4)
        eq(await c.get(key), b'a1a2')
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.append, key, 'g')

    async def test_bitcount(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.bitcount(key), 0)
        eq(await c.setbit(key, 5, 1), 0)
        eq(await c.bitcount(key), 1)
        eq(await c.setbit(key, 6, 1), 0)
        eq(await c.bitcount(key), 2)
        eq(await c.setbit(key, 5, 0), 1)
        eq(await c.bitcount(key), 1)
        eq(await c.setbit(key, 9, 1), 0)
        eq(await c.setbit(key, 17, 1), 0)
        eq(await c.setbit(key, 25, 1), 0)
        eq(await c.setbit(key, 33, 1), 0)
        eq(await c.bitcount(key), 5)
        eq(await c.bitcount(key, 0, -1), 5)
        eq(await c.bitcount(key, 2, 3), 2)
        eq(await c.bitcount(key, 2, -1), 3)
        eq(await c.bitcount(key, -2, -1), 2)
        eq(await c.bitcount(key, 1, 1), 1)
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.bitcount, key)

    async def test_bitop_not_empty_string(self):
        key = self.randomkey()
        des = key + 'd'
        c = self.client
        eq = self.assertEqual
        eq(await c.set(key, ''), True)
        eq(await c.bitop('not', des, key), 0)
        eq(await c.get(des), None)

    async def test_bitop_not(self):
        key = self.randomkey()
        des = key + 'd'
        c = self.client
        eq = self.assertEqual
        test_str = b'\xAA\x00\xFF\x55'
        correct = ~0xAA00FF55 & 0xFFFFFFFF
        eq(await c.set(key, test_str), True)
        eq(await c.bitop('not', des, key), 4)
        result = await c.get(des)
        self.assertEqual(int(binascii.hexlify(result), 16), correct)

    async def test_bitop_not_in_place(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        test_str = b'\xAA\x00\xFF\x55'
        correct = ~0xAA00FF55 & 0xFFFFFFFF
        eq(await c.set(key, test_str), True)
        eq(await c.bitop('not', key, key), 4)
        result = await c.get(key)
        assert int(binascii.hexlify(result), 16) == correct

    async def test_bitop_single_string(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        test_str = b'\x01\x02\xFF'
        eq(await c.set(key, test_str), True)
        eq(await c.bitop('and', key+'1', key), 3)
        eq(await c.bitop('or', key+'2', key), 3)
        eq(await c.bitop('xor', key+'3', key), 3)
        eq(await c.get(key + '1'), test_str)
        eq(await c.get(key + '2'), test_str)
        eq(await c.get(key + '3'), test_str)

    async def test_bitop_string_operands(self):
        c = self.client
        eq = self.assertEqual
        key1 = self.randomkey()
        key2 = key1 + '2'
        des1 = key1 + 'd1'
        des2 = key1 + 'd2'
        des3 = key1 + 'd3'
        eq(await c.set(key1, b'\x01\x02\xFF\xFF'), True)
        eq(await c.set(key2, b'\x01\x02\xFF'), True)
        eq(await c.bitop('and', des1, key1, key2), 4)
        eq(await c.bitop('or', des2, key1, key2), 4)
        eq(await c.bitop('xor', des3, key1, key2), 4)
        res1 = await c.get(des1)
        res2 = await c.get(des2)
        res3 = await c.get(des3)
        self.assertEqual(int(binascii.hexlify(res1), 16), 0x0102FF00)
        self.assertEqual(int(binascii.hexlify(res2), 16), 0x0102FFFF)
        self.assertEqual(int(binascii.hexlify(res3), 16), 0x000000FF)

    async def test_decr(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.decr(key), -1)
        eq(await c.get(key), b'-1')
        eq(await c.decr(key), -2)
        eq(await c.get(key), b'-2')
        eq(await c.decr(key, 5), -7)
        eq(await c.get(key), b'-7')

    async def test_getbit(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.getbit(key, 5), 0)
        eq(await c.setbit(key, 5, 1), 0)
        eq(await c.getbit(key, 5), 1)
        await self.wait.assertRaises(ResponseError, c.getbit, key, -1)
        eq(await c.getbit(key, 4), 0)
        eq(await c.setbit(key, 4, 1), 0)
        # set bit 4
        eq(await c.getbit(key, 4), 1)
        eq(await c.getbit(key, 5), 1)
        # set bit 5 again
        eq(await c.setbit(key, 5, 1), 1)
        eq(await c.getbit(key, 5), 1)
        #
        eq(await c.getbit(key, 30), 0)
        #
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.getbit, key, 1)

    async def test_getrange(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.getrange(key, 0, 0), b'')
        eq(await c.set(key, 'Hello there'), True)
        eq(await c.getrange(key, 0, 0), b'H')
        eq(await c.getrange(key, 0, 4), b'Hello')
        eq(await c.getrange(key, 5, 5), b' ')
        eq(await c.getrange(key, 20, 25), b'')
        eq(await c.getrange(key, -5, -1), b'there')
        await self.wait.assertRaises(ResponseError, c.getrange, key, 1, 'b')
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.getrange, key, 1, 2)

    async def test_getset(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.getset(key, 'foo'), None)
        eq(await c.getset(key, 'bar'), b'foo')
        eq(await c.get(key), b'bar')

    async def test_get(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        await c.set(key, 'foo')
        eq(await c.get(key), b'foo')
        eq(await c.get('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'), None)

    async def test_incr(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.incr(key), 1)
        eq(await c.get(key), b'1')
        eq(await c.incr(key), 2)
        eq(await c.get(key), b'2')
        eq(await c.incr(key, 5), 7)
        eq(await c.get(key), b'7')

    async def test_incrby(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.incrby(key), 1)
        eq(await c.incrby(key, 4), 5)
        eq(await c.get(key), b'5')

    async def test_incrbyfloat(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.incrbyfloat(key), 1.0)
        eq(await c.get(key), b'1')
        eq(await c.incrbyfloat(key, 1.1), 2.1)
        eq(await c.get(key), b'2.1')

    async def test_mget(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        key3 = key2 + 'y'
        eq = self.assertEqual
        c = self.client
        eq(await c.set(key1, 'foox'), True)
        eq(await c.set(key2, 'fooxx'), True)
        eq(await c.mget(key1, key2, key3), [b'foox', b'fooxx', None])

    async def test_msetnx(self):
        key1 = self.randomkey()
        key2 = self.randomkey()
        key3 = self.randomkey()
        key4 = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.msetnx(key1, '1', key2, '2', key3, '3'), True)
        eq(await c.msetnx(key1, 'x', key4, 'y'), False)
        values = await c.mget(key1, key2, key3, key4)
        self.assertEqual(len(values), 4)
        for value, target in zip(values, (b'1', b'2', b'3', None)):
            self.assertEqual(value, target)

    async def test_set_ex(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.set(key, '1', ex=5), True)
        ttl = await c.ttl(key)
        self.assertTrue(0 < ttl <= 5)

    async def test_set_ex_timedelta(self):
        expire_at = datetime.timedelta(seconds=5)
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.set(key, '1', ex=expire_at), True)
        ttl = await c.ttl(key)
        self.assertTrue(0 < ttl <= 5)

    async def test_set_nx(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.set(key, '1', nx=True), True)
        eq(await c.set(key, '2', nx=True), False)
        eq(await c.get(key), b'1')

    async def test_set_px(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.set(key, '1', px=10000, nx=True), True)
        eq(await c.get(key), b'1')
        pttl = await c.pttl(key)
        self.assertTrue(0 < pttl <= 10000)
        ttl = await c.ttl(key)
        self.assertTrue(0 < ttl <= 10)

    async def test_set_px_timedelta(self):
        expire_at = datetime.timedelta(milliseconds=5000)
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.set(key, '1', px=expire_at, nx=True), True)
        eq(await c.get(key), b'1')
        pttl = await c.pttl(key)
        self.assertTrue(0 < pttl <= 5000)
        ttl = await c.ttl(key)
        self.assertTrue(0 < ttl <= 5)

    async def test_set_xx(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.set(key, '1', xx=True), False)
        eq(await c.get(key), None)
        eq(await c.set(key, 'bar'), True)
        eq(await c.set(key, '2', xx=True), True)
        eq(await c.get(key), b'2')

    async def test_setrange(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.setrange(key, 5, 'foo'), 8)
        eq(await c.get(key), b'\0\0\0\0\0foo')
        eq(await c.set(key, 'abcdefghijh'), True)
        eq(await c.setrange(key, 6, '12345'), 11)
        eq(await c.get(key), b'abcdef12345')

    async def test_strlen(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.strlen(key), 0)
        eq(await c.set(key, 'foo'), True)
        eq(await c.strlen(key), 3)

    ###########################################################################
    #    HASHES
    async def test_hdel(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.hdel(key, 'f1', 'f2', 'gh'), 0)
        eq(await c.hmset(key, {'f1': 1, 'f2': 'hello', 'f3': 'foo'}), True)
        eq(await c.hdel(key, 'f1', 'f2', 'gh'), 2)
        eq(await c.hdel(key, 'fgf'), 0)
        eq(await c.type(key), 'hash')
        eq(await c.hdel(key, 'f3'), 1)
        eq(await c.type(key), 'none')
        await self._remove_and_push(key, 0)
        await self.wait.assertRaises(ResponseError, c.hdel, key, 'foo')

    async def test_hexists(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.hexists(key, 'foo'), False)
        eq(await c.hmset(key, {'f1': 1, 'f2': 'hello', 'f3': 'foo'}), True)
        eq(await c.hexists(key, 'f3'), True)
        eq(await c.hexists(key, 'f5'), False)
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.hexists, key, 'foo')

    async def test_hset_hget(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.hget(key, 'foo'), None)
        eq(await c.hset(key, 'foo', 4), 1)
        eq(await c.hget(key, 'foo'), b'4')
        eq(await c.hset(key, 'foo', 6), 0)
        eq(await c.hget(key, 'foo'), b'6')
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.hset, key, 'foo', 7)
        await self.wait.assertRaises(ResponseError, c.hget, key, 'foo')
        await self.wait.assertRaises(ResponseError, c.hmset, key, 'foo')

    async def test_hgetall(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        h = {b'f1': b'1', b'f2': b'hello', b'f3': b'foo'}
        eq(await c.hgetall(key), {})
        eq(await c.hmset(key, h), True)
        eq(await c.hgetall(key), h)
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.hgetall, key)

    async def test_hincrby(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.hincrby(key, 'foo', 1), 1)
        eq(await c.hincrby(key, 'foo', 2), 3)
        eq(await c.hincrby(key, 'foo', -1), 2)
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.hincrby, key, 'foo', 3)

    async def test_hincrbyfloat(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.hincrbyfloat(key, 'foo', 1), 1.0)
        eq(await c.hincrbyfloat(key, 'foo', 2.5), 3.5)
        eq(await c.hincrbyfloat(key, 'foo', -1.1), 2.4)
        await self._remove_and_push(key)

    async def test_hkeys_hvals_hlen_hmget(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        h = {b'f1': b'1', b'f2': b'hello', b'f3': b'foo'}
        eq(await c.hkeys(key), [])
        eq(await c.hvals(key), [])
        eq(await c.hlen(key), 0)
        eq(await c.hmset(key, h), True)
        keys = await c.hkeys(key)
        vals = await c.hvals(key)
        self.assertEqual(sorted(keys), sorted(h))
        self.assertEqual(sorted(vals), sorted(h.values()))
        eq(await c.hlen(key), 3)
        eq(await c.hmget(key, 'f1', 'f3', 'hj'),
           {'f1': b'1', 'f3': b'foo', 'hj': None})
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.hkeys, key)
        await self.wait.assertRaises(ResponseError, c.hvals, key)
        await self.wait.assertRaises(ResponseError, c.hlen, key)
        await self.wait.assertRaises(ResponseError, c.hmget, key, 'f1', 'f2')

    async def test_hsetnx(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.hsetnx(key, 'a', 'foo'), 1)
        eq(await c.hget(key, 'a'), b'foo')
        eq(await c.hsetnx(key, 'a', 'bla'), 0)
        eq(await c.hget(key, 'a'), b'foo')
        await self._remove_and_push(key)
        await self.wait.assertRaises(ResponseError, c.hsetnx, key, 'a', 'jk')

    ###########################################################################
    #    LISTS
    async def test_blpop(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        bk1 = key1.encode('utf-8')
        bk2 = key2.encode('utf-8')
        eq = self.assertEqual
        c = self.client
        await self.wait.assertRaises(ResponseError, c.blpop, key1, 'bla')
        eq(await c.rpush(key1, 1, 2), 2)
        eq(await c.rpush(key2, 3, 4), 2)
        eq(await c.blpop((key2, key1), 1), (bk2, b'3'))
        eq(await c.blpop((key2, key1), 1), (bk2, b'4'))
        eq(await c.blpop((key2, key1), 1), (bk1, b'1'))
        eq(await c.blpop((key2, key1), 1), (bk1, b'2'))
        eq(await c.blpop((key2, key1), 1), None)
        eq(await c.rpush(key1, '1'), 1)
        eq(await c.blpop(key1, 1), (bk1, b'1'))

    async def test_brpop(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        bk1 = key1.encode('utf-8')
        bk2 = key2.encode('utf-8')
        eq = self.assertEqual
        c = self.client
        eq(await c.rpush(key1, 1, 2), 2)
        eq(await c.rpush(key2, 3, 4), 2)
        eq(await c.brpop((key2, key1), 1), (bk2, b'4'))
        eq(await c.brpop((key2, key1), 1), (bk2, b'3'))
        eq(await c.brpop((key2, key1), 1), (bk1, b'2'))
        eq(await c.brpop((key2, key1), 1), (bk1, b'1'))
        eq(await c.brpop((key2, key1), 1), None)
        eq(await c.rpush(key1, '1'), 1)
        eq(await c.brpop(key1, 1), (bk1, b'1'))

    async def test_brpoplpush(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        eq = self.assertEqual
        c = self.client
        eq(await c.rpush(key1, 1, 2), 2)
        eq(await c.rpush(key2, 3, 4), 2)
        eq(await c.brpoplpush(key1, key2), b'2')
        eq(await c.brpoplpush(key1, key2), b'1')
        eq(await c.brpoplpush(key1, key2, timeout=1), None)
        eq(await c.lrange(key1, 0, -1), [])
        eq(await c.lrange(key2, 0, -1), [b'1', b'2', b'3', b'4'])

    async def test_brpoplpush_empty_string(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        eq = self.assertEqual
        c = self.client
        eq(await c.rpush(key1, ''), 1)
        eq(await c.brpoplpush(key1, key2), b'')

    async def test_lindex_llen(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.lindex(key, '0'), None)
        eq(await c.llen(key), 0)
        eq(await c.rpush(key, '1', '2', '3'), 3)
        eq(await c.lindex(key, '0'), b'1')
        eq(await c.lindex(key, '1'), b'2')
        eq(await c.lindex(key, '2'), b'3')
        eq(await c.lindex(key, '3'), None)
        eq(await c.llen(key), 3)
        await self._remove_and_sadd(key)
        await self.wait.assertRaises(ResponseError, c.lindex, key, '1')
        await self.wait.assertRaises(ResponseError, c.llen, key)

    async def test_linsert(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.linsert(key, 'after', '2', '2.5'), 0)
        eq(await c.rpush(key, '1', '2', '3'), 3)
        eq(await c.linsert(key, 'after', '2', '2.5'), 4)
        eq(await c.lrange(key, 0, -1), [b'1', b'2', b'2.5', b'3'])
        eq(await c.linsert(key, 'before', '2', '1.5'), 5)
        eq(await c.lrange(key, 0, -1), [b'1', b'1.5', b'2', b'2.5', b'3'])
        eq(await c.linsert(key, 'before', '100', '1.5'), -1)
        await self.wait.assertRaises(ResponseError, c.linsert, key,
                                     'banana', '2', '2.5')
        await self._remove_and_sadd(key)
        await self.wait.assertRaises(ResponseError, c.linsert, key,
                                     'after', '2', '2.5')

    async def test_lpop(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.lpop(key), None)
        eq(await c.rpush(key, 1, 2), 2)
        eq(await c.lpop(key), b'1')
        eq(await c.lpop(key), b'2')
        eq(await c.lpop(key), None)
        eq(await c.type(key), 'none')
        await self._remove_and_sadd(key, 0)
        await self.wait.assertRaises(ResponseError, c.lpop, key)
        await self.wait.assertRaises(ResponseError, c.lpush, key, 4)

    async def test_lset(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.rpush(key, '1', '2', '3'), 3)
        eq(await c.lrange(key, 0, -1), [b'1', b'2', b'3'])
        eq(await c.lset(key, 1, '4'), True)
        eq(await c.lrange(key, 0, 2), [b'1', b'4', b'3'])

    async def test_ltrim(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        eq(await c.rpush(key, '1', '2', '3'), 3)
        eq(await c.ltrim(key, 0, 1), True)
        eq(await c.lrange(key, 0, -1), [b'1', b'2'])

    async def test_lpushx_rpushx(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.lpushx(key, 'b'), 0)
        eq(await c.rpushx(key, 'b'), 0)
        eq(await c.lrange(key, 0, -1), [])
        eq(await c.lpush(key, 'a'), 1)
        eq(await c.lpushx(key, 'b'), 2)
        eq(await c.rpushx(key, 'c'), 3)
        eq(await c.lrange(key, 0, -1), [b'b', b'a', b'c'])
        await self._remove_and_sadd(key)
        await self.wait.assertRaises(ResponseError, c.lpushx, key, 'g')

    async def test_lrem(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.lrem(key, 1, 'a'), 0)
        eq(await c.rpush(key, 'a', 'a', 'a', 'b', 'a', 'a'), 6)
        eq(await c.lrem(key, 1, 'a'), 1)
        eq(await c.lrange(key, 0, -1), [b'a', b'a', b'b', b'a', b'a'])
        eq(await c.lrem(key, -1, 'a'), 1)
        eq(await c.lrange(key, 0, -1), [b'a', b'a', b'b', b'a'])
        eq(await c.lrem(key, 0, 'a'), 3)
        await self.wait.assertRaises(ResponseError, c.lrem, key, 'g', 'foo')
        eq(await c.lrange(key, 0, -1), [b'b'])
        eq(await c.lrem(key, 0, 'b'), 1)
        await self._remove_and_sadd(key, 0)
        await self.wait.assertRaises(ResponseError, c.lrem, key, 1)

    async def test_rpop(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.rpop(key), None)
        eq(await c.rpush(key, 1, 2), 2)
        eq(await c.rpop(key), b'2')
        eq(await c.rpop(key), b'1')
        eq(await c.rpop(key), None)
        eq(await c.type(key), 'none')

    async def test_rpoplpush(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        eq = self.assertEqual
        c = self.client
        eq(await c.rpush(key1, 'a1', 'a2', 'a3'), 3)
        eq(await c.rpush(key2, 'b1', 'b2', 'b3'), 3)
        eq(await c.rpoplpush(key1, key2), b'a3')
        eq(await c.lrange(key1, 0, -1), [b'a1', b'a2'])
        eq(await c.lrange(key2, 0, -1), [b'a3', b'b1', b'b2', b'b3'])

    ###########################################################################
    #    SORT
    async def test_sort_basic(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        await c.rpush(key, '3', '2', '1', '4')
        eq(await c.sort(key), [b'1', b'2', b'3', b'4'])

    async def test_sort_limited(self):
        key = self.randomkey()
        c = self.client
        eq = self.assertEqual
        await c.rpush(key, '3', '2', '1', '4')
        eq(await c.sort(key, start=1, num=2), [b'2', b'3'])

    async def test_sort_by(self):
        key = self.randomkey()
        key2 = self.randomkey()
        c = self.client
        eq = self.assertEqual
        pipe = c.pipeline()
        pipe.mset('%s:1' % key, 8,
                  '%s:2' % key, 3,
                  '%s:3' % key, 5)
        pipe.rpush(key2, '3', '2', '1')
        res = await pipe.commit()
        self.assertEqual(len(res), 2)
        eq(await c.sort(key2, by='%s:*' % key), [b'2', b'3', b'1'])

    async def test_sort_get(self):
        key = self.randomkey()
        key2 = self.randomkey()
        c = self.client
        eq = self.assertEqual
        pipe = c.pipeline()
        pipe.mset('%s:1' % key, 'u1',
                  '%s:2' % key, 'u2',
                  '%s:3' % key, 'u3')
        pipe.rpush(key2, '3', '2', '1')
        res = await pipe.commit()
        self.assertEqual(len(res), 2)
        eq(await c.sort(key2, get='%s:*' % key), [b'u1', b'u2', b'u3'])

    async def test_sort_get_multi(self):
        key = self.randomkey()
        key2 = self.randomkey()
        c = self.client
        eq = self.assertEqual
        pipe = c.pipeline()
        pipe.mset('%s:1' % key, 'u1',
                  '%s:2' % key, 'u2',
                  '%s:3' % key, 'u3')
        pipe.rpush(key2, '3', '2', '1')
        res = await pipe.commit()
        self.assertEqual(len(res), 2)
        eq(await c.sort(key2, get=('%s:*' % key, '#')),
           [b'u1', b'1', b'u2', b'2', b'u3', b'3'])
        eq(await c.sort(key2, get=('%s:*' % key, '#'), groups=True),
           [(b'u1', b'1'), (b'u2', b'2'), (b'u3', b'3')])

    ###########################################################################
    #    SETS
    async def test_sadd_scard(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        members = (b'1', b'2', b'3', b'2')
        eq(await c.sadd(key, *members), 3)
        eq(await c.smembers(key), set(members))
        eq(await c.scard(key), 3)

    async def test_sdiff(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        eq(await c.sdiff(key, key2), set((b'1', b'2', b'3')))
        eq(await c.sadd(key2, 2, 3), 2)
        eq(await c.sdiff(key, key2), set([b'1']))

    async def test_sdiffstore(self):
        key = self.randomkey()
        key2 = key + '2'
        des = key + 'd'
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        eq(await c.sdiffstore(des, key, key2), 3)
        eq(await c.smembers(des), set([b'1', b'2', b'3']))
        eq(await c.sadd(key2, 2, 3), 2)
        eq(await c.sdiffstore(des, key, key2), 1)
        eq(await c.smembers(des), set([b'1']))

    async def test_sinter(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        eq(await c.sinter(key, key2), set())
        eq(await c.sadd(key2, 2, 3), 2)
        eq(await c.sinter(key, key2), set([b'2', b'3']))

    async def test_sinterstore(self):
        key = self.randomkey()
        key2 = key + '2'
        des = key + 'd'
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        eq(await c.sinterstore(des, key, key2), 0)
        eq(await c.smembers(des), set())
        eq(await c.sadd(key2, 2, 3), 2)
        eq(await c.sinterstore(des, key, key2), 2)
        eq(await c.smembers(des), set([b'2', b'3']))

    async def test_sismember(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        eq(await c.sismember(key, 1), True)
        eq(await c.sismember(key, 2), True)
        eq(await c.sismember(key, 3), True)
        eq(await c.sismember(key, 4), False)

    async def test_smove(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.assertEqual
        c = self.client
        eq(await c.smove(key, key2, 1), False)
        eq(await c.sadd(key, 1, 2), 2)
        eq(await c.sadd(key2, 3, 4), 2)
        eq(await c.smove(key, key2, 1), True)
        eq(await c.smembers(key), set([b'2']))
        eq(await c.smembers(key2), set([b'1', b'3', b'4']))

    async def test_spop(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        value = await c.spop(key)
        self.assertTrue(value in set([b'1', b'2', b'3']))
        eq(await c.smembers(key), set([b'1', b'2', b'3']) - set([value]))

    async def test_srandmember(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        value = await c.srandmember(key)
        self.assertTrue(value in set((b'1', b'2', b'3')))
        eq(await c.smembers(key), set((b'1', b'2', b'3')))

    async def test_srandmember_multi_value(self):
        s = [b'1', b'2', b'3']
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, *s), 3)
        randoms = await c.srandmember(key, 2)
        self.assertEqual(len(randoms), 2)
        self.assertEqual(set(randoms).intersection(s), set(randoms))

    async def test_srem(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3, 4), 4)
        eq(await c.srem(key, 5), 0)
        eq(await c.srem(key, 5), 0)
        eq(await c.srem(key, 2, 4), 2)
        eq(await c.smembers(key), set([b'1', b'3']))

    async def test_sunion(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        eq(await c.sunion(key, key2), set((b'1', b'2', b'3')))
        eq(await c.sadd(key2, 2, 3, 4), 3)
        eq(await c.sunion(key, key2), set((b'1', b'2', b'3', b'4')))

    async def test_sunionstore(self):
        key = self.randomkey()
        key2 = key + '2'
        des = key + 'd'
        eq = self.assertEqual
        c = self.client
        eq(await c.sadd(key, 1, 2, 3), 3)
        eq(await c.sunionstore(des, key, key2), 3)
        eq(await c.smembers(des), set([b'1', b'2', b'3']))
        eq(await c.sadd(key2, 2, 3, 4), 3)
        eq(await c.sunionstore(des, key, key2), 4)
        eq(await c.smembers(des), set([b'1', b'2', b'3', b'4']))

    ###########################################################################
    #    SORTED SETS
    async def test_zadd_zcard(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3), 3)
        eq(await c.zrange(key, 0, -1), [b'a1', b'a2', b'a3'])
        eq(await c.zcard(key), 3)

    async def test_zcount(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3), 3)
        eq(await c.zcount(key, '-inf', '+inf'), 3)
        eq(await c.zcount(key, '(1', 2), 1)
        eq(await c.zcount(key, '(1', '(3'), 1)
        eq(await c.zcount(key, 1, 3), 3)

    async def test_zincrby(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3), 3)
        eq(await c.zincrby(key, 1, 'a2'), 3.0)
        eq(await c.zincrby(key, 5, 'a3'), 8.0)
        eq(await c.zscore(key, 'a2'), 3.0)
        eq(await c.zscore(key, 'a3'), 8.0)
        eq(await c.zscore(key, 'blaaa'), None)

    async def test_zinterstore_sum(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key1, a1=1, a2=2, a3=1), 3)
        eq(await c.zadd(key2, a1=2, a2=2, a3=2), 3)
        eq(await c.zadd(key3, a1=6, a3=5, a4=4), 3)
        eq(await c.zinterstore(des, (key1, key2, key3)), 2)
        eq(await c.zrange(des, 0, -1, withscores=True),
           Zset(((8.0, b'a3'), (9.0, b'a1'))))

    async def test_zinterstore_max(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key1, a1=1, a2=2, a3=1), 3)
        eq(await c.zadd(key2, a1=2, a2=2, a3=2), 3)
        eq(await c.zadd(key3, a1=6, a3=5, a4=4), 3)
        eq(await c.zinterstore(des, (key1, key2, key3), aggregate='max'), 2)
        eq(await c.zrange(des, 0, -1, withscores=True),
           Zset(((5.0, b'a3'), (6.0, b'a1'))))

    async def test_zinterstore_min(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key1, a1=1, a2=2, a3=1), 3)
        eq(await c.zadd(key2, a1=2, a2=2, a3=2), 3)
        eq(await c.zadd(key3, a1=6, a3=5, a4=4), 3)
        eq(await c.zinterstore(des, (key1, key2, key3), aggregate='min'), 2)
        eq(await c.zrange(des, 0, -1, withscores=True),
           Zset(((1.0, b'a3'), (1.0, b'a1'))))

    async def test_zinterstore_with_weights(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key1, a1=1, a2=2, a3=1), 3)
        eq(await c.zadd(key2, a1=2, a2=2, a3=2), 3)
        eq(await c.zadd(key3, a1=6, a3=5, a4=4), 3)
        eq(await c.zinterstore(des, (key1, key2, key3), weights=(1, 2, 3)), 2)
        eq(await c.zrange(des, 0, -1, withscores=True),
           Zset(((20.0, b'a3'), (23.0, b'a1'))))

    async def test_zrange(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3), 3)
        eq(await c.zrange(key, 0, 1), [b'a1', b'a2'])
        eq(await c.zrange(key, 1, 2), [b'a2', b'a3'])
        eq(await c.zrange(key, 0, 1, withscores=True),
           Zset([(1, b'a1'), (2, b'a2')]))
        eq(await c.zrange(key, 1, 2, withscores=True),
           Zset([(2, b'a2'), (3, b'a3')]))

    async def test_zrangebyscore(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        eq(await c.zrangebyscore(key, 2, 4), [b'a2', b'a3', b'a4'])
        # slicing with start/num
        eq(await c.zrangebyscore(key, 2, 4, offset=1, count=2), [b'a3', b'a4'])
        # withscores
        eq(await c.zrangebyscore(key, 2, 4, withscores=True),
           Zset([(2.0, b'a2'), (3.0, b'a3'), (4.0, b'a4')]))

    async def test_zrank(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        eq(await c.zrank(key, 'a1'), 0)
        eq(await c.zrank(key, 'a2'), 1)
        eq(await c.zrank(key, 'a6'), None)

    async def test_zrem(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        eq(await c.zrem(key, 'a2'), 1)
        eq(await c.zrange(key, 0, -1), [b'a1', b'a3', b'a4', b'a5'])
        eq(await c.zrem(key, 'b'), 0)
        eq(await c.zrange(key, 0, -1), [b'a1', b'a3', b'a4', b'a5'])
        eq(await c.zrem(key, 'a3', 'a5', 'h'), 2)
        eq(await c.zrange(key, 0, -1), [b'a1', b'a4'])
        eq(await c.zrem(key, 'a1', 'a4'), 2)
        eq(await c.type(key), 'none')

    async def test_zremrangebyrank(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        eq(await c.zremrangebyrank(key, 1, 3), 3)
        eq(await c.zrange(key, 0, 5), [b'a1', b'a5'])

    async def test_zremrangebyscore(self):
        key = self.randomkey()
        eq = self.assertEqual
        c = self.client
        eq(await c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        eq(await c.zremrangebyscore(key, 2, 4), 3)
        eq(await c.zrange(key, 0, -1), [b'a1', b'a5'])
        eq(await c.zremrangebyscore(key, 2, 4), 0)
        eq(await c.zrange(key, 0, -1), [b'a1', b'a5'])

    ###########################################################################
    #    CONNECTION
    async def test_ping(self):
        result = await self.client.ping()
        self.assertTrue(result)

    async def test_echo(self):
        result = await self.client.echo('Hello')
        self.assertEqual(result, b'Hello')

    ###########################################################################
    #    SERVER
    async def test_dbsize(self):
        await self.client.set('one_at_least', 'foo')
        result = await self.client.dbsize()
        self.assertTrue(result >= 1)

    async def test_info(self):
        info = await self.client.info()
        self.assertTrue(info)
        self.assertIsInstance(info, dict)

    async def test_time(self):
        t = await self.client.time()
        self.assertIsInstance(t, tuple)
        total = t[0] + 0.000001*t[1]
        self.assertTrue(total)

    ###########################################################################
    #    PUBSUB
    def test_handler(self):
        client = self.client
        pubsub = client.pubsub()
        self.assertEqual(client.store, pubsub.store)
        self.assertEqual(client.store._loop, pubsub._loop)
        self.assertEqual(pubsub._connection, None)

    async def test_subscribe_one(self):
        key = self.randomkey()
        pubsub1 = self.client.pubsub()
        self.assertFalse(pubsub1._connection)
        # Subscribe to one channel
        await pubsub1.subscribe(key)
        count = await pubsub1.count(key)
        self.assertEqual(len(count), 1)
        self.assertEqual(count[key.encode('utf-8')], 1)
        #
        pubsub2 = self.client.pubsub()
        await pubsub2.subscribe(key)
        count = await pubsub1.count(key)
        self.assertEqual(len(count), 1)
        self.assertEqual(count[key.encode('utf-8')], 2)

    async def test_subscribe_many(self):
        base = self.randomkey()
        key1 = base + '_a'
        key2 = base + '_b'
        key3 = base + '_c'
        key4 = base + 'x'
        pubsub = self.client.pubsub()
        await pubsub.subscribe(key1, key2, key3, key4)
        channels = await pubsub.channels(base + '_*')
        self.assertEqual(len(channels), 3)
        count = await pubsub.count(key1, key2, key3)
        self.assertEqual(len(count), 3)
        self.assertEqual(count[key1.encode('utf-8')], 1)
        self.assertEqual(count[key2.encode('utf-8')], 1)
        self.assertEqual(count[key3.encode('utf-8')], 1)

    async def test_publish(self):
        pubsub = self.client.pubsub()
        listener = Listener()
        pubsub.add_client(listener)
        await pubsub.subscribe('chat')
        result = await pubsub.publish('chat', 'Hello')
        self.assertTrue(result >= 0)
        channel, message = await listener.get()
        self.assertEqual(channel, 'chat')
        self.assertEqual(message, b'Hello')

    async def test_pattern_subscribe(self):
        # switched off for redis. Issue #95
        if self.store.name == 'pulsar':
            eq = self.assertEqual
            pubsub = self.client.pubsub(protocol=StringProtocol())
            listener = Listener()
            pubsub.add_client(listener)
            eq(await pubsub.psubscribe('f*'), None)
            eq(await pubsub.publish('foo', 'hello foo'), 1)
            channel, message = await listener.get()
            self.assertEqual(channel, 'foo')
            self.assertEqual(message, 'hello foo')
            eq(await pubsub.punsubscribe(), None)
            # await listener.get()

    ###########################################################################
    #    TRANSACTION
    async def test_watch(self):
        key1 = self.randomkey()
        # key2 = key1 + '2'
        result = await self.client.watch(key1)
        self.assertEqual(result, 1)


class TestPulsarStore(RedisCommands, unittest.TestCase):
    app_cfg = None

    @classmethod
    async def setUpClass(cls):
        server = PulsarDS(name=cls.__name__.lower(),
                          bind='127.0.0.1:0',
                          concurrency=cls.cfg.concurrency,
                          redis_py_parser=cls.redis_py_parser)
        cls.app_cfg = await pulsar.send('arbiter', 'run', server)
        cls.pulsards_uri = 'pulsar://%s:%s' % cls.app_cfg.addresses[0]
        cls.store = cls.create_store('%s/9' % cls.pulsards_uri)
        cls.client = cls.store.client()

    @classmethod
    def tearDownClass(cls):
        if cls.app_cfg is not None:
            return pulsar.send('arbiter', 'kill_actor', cls.app_cfg.name)

    def test_store_methods(self):
        store = self.create_store('%s/8' % self.pulsards_uri)
        self.assertEqual(store.database, 8)
        store.database = 10
        self.assertEqual(store.database, 10)
        self.assertTrue(store.dns.startswith('%s/10?' % self.pulsards_uri))
        self.assertEqual(store.encoding, 'utf-8')
        self.assertTrue(repr(store))


@unittest.skipUnless(pulsar.HAS_C_EXTENSIONS, 'Requires cython extensions')
class TestPulsarStorePyParser(TestPulsarStore):
    redis_py_parser = True



import unittest


class TestMixins(unittest.TestCase):

    def testLocal(self):
        from pulsar.utils.structures import AttributeDictionary
        from pulsar.utils.log import LocalMixin
        elem = LocalMixin()
        el = elem.local
        self.assertTrue(isinstance(el, AttributeDictionary))
        self.assertEqual(id(elem.local), id(el))
        self.assertEqual(elem.local.process, None)
        elem.local.process = True
        self.assertEqual(elem.local.process, True)



import unittest
import asyncio

from pulsar import Lock


class TestLock(unittest.TestCase):

    def test_lock_init(self):
        lock = Lock('foo')
        self.assertEqual(lock.name, 'foo')
        lock2 = Lock('foo')
        self.assertEqual(lock2.name, 'foo')
        self.assertNotEqual(lock, lock2)
        self.assertEqual(lock._lock, lock2._lock)

    async def test_lock(self):
        lock1 = Lock('test', blocking=False)
        lock2 = Lock('test', blocking=False)
        self.assertEqual(await lock1.acquire(), True)
        self.assertEqual(await lock2.acquire(), False)
        self.assertFalse(lock2.locked())
        await lock1.release()
        self.assertFalse(lock1.locked())

    async def test_lock_blocking(self):
        lock1 = Lock('test1')
        lock2 = Lock('test1', blocking=1)
        self.assertEqual(await lock1.acquire(), True)
        start = lock2._loop.time()
        self.assertEqual(await lock2.acquire(), False)
        self.assertGreaterEqual(lock2._loop.time() - start, 1)
        self.assertFalse(lock2.locked())
        await lock1.release()
        self.assertFalse(lock1.locked())

    async def test_lock_timeout(self):
        lock = Lock('test2', timeout=1)
        self.assertEqual(await lock.acquire(), True)
        await asyncio.sleep(1.5)
        self.assertFalse(lock.locked())

    async def test_lock_timeout_lock(self):
        lock1 = Lock('test3', timeout=1)
        lock2 = Lock('test3', blocking=True)
        self.assertEqual(await lock1.acquire(), True)
        self.assertTrue(lock1.locked())
        future = asyncio.ensure_future(lock2.acquire())
        await asyncio.sleep(1.5)
        self.assertFalse(lock1.locked())
        await future
        self.assertTrue(lock2.locked())

    async def test_context(self):
        lock = Lock('test4', blocking=1)
        async with Lock('test4'):
            self.assertEqual(await lock.acquire(), False)
            self.assertFalse(lock.locked())
        async with lock:
            self.assertTrue(lock.locked())
        self.assertFalse(lock.locked())



import asyncio

from pulsar import ensure_future
from pulsar.apps.data import LockError


class RedisLockTests:

    async def _release(self, lock, time):
        await asyncio.sleep(time)
        self.assertEqual(await lock.release(), True)

    async def test_lock(self):
        key = self.randomkey()
        eq = self.assertEqual
        lock = self.client.lock(key)
        self.assertEqual(lock.name, key)
        self.assertFalse(lock._token)

        eq(await lock.acquire(), True)
        self.assertTrue(lock._token)
        eq(await self.client.get(key), lock._token)
        # assert sr.ttl('foo') == -1
        eq(await lock.release(), True)
        eq(await self.client.get(key), None)

    async def test_competing_locks(self):
        key = self.randomkey()
        eq = self.assertEqual
        lock1 = self.client.lock(key, blocking=False)
        lock2 = self.client.lock(key, blocking=False)
        eq(await lock1.acquire(), True)
        eq(await lock2.acquire(), False)
        eq(await lock1.release(), True)
        eq(await lock2.acquire(), True)
        eq(await lock1.acquire(), False)
        eq(await lock2.release(), True)

    async def test_timeout(self):
        key = self.randomkey()
        eq = self.assertEqual
        lock = self.client.lock(key, timeout=10)
        eq(await lock.acquire(), True)
        ttl = await self.client.ttl(lock.name)
        self.assertTrue(2 < ttl <= 10)
        eq(await lock.release(), True)

    async def test_float_timeout(self):
        key = self.randomkey()
        eq = self.assertEqual
        lock = self.client.lock(key, timeout=9.5)
        eq(await lock.acquire(), True)
        ttl = await self.client.pttl(lock.name)
        self.assertTrue(4000 < ttl <= 9500)
        eq(await lock.release(), True)

    async def test_blocking_timeout(self):
        key = self.randomkey()
        eq = self.assertEqual
        lock1 = self.client.lock(key)
        lock2 = self.client.lock(key, blocking=0.5)
        eq(await lock1.acquire(), True)
        start = lock1._loop.time()
        eq(await lock2.acquire(), False)
        self.assertTrue(lock1._loop.time() - start > 0.5)
        eq(await lock1.release(), True)

    async def test_blocking_timeout_acquire(self):
        key = self.randomkey()
        eq = self.assertEqual
        lock1 = self.client.lock(key)
        lock2 = self.client.lock(key, blocking=5)
        eq(await lock1.acquire(), True)
        ensure_future(self._release(lock1, 0.5))
        start = lock2._loop.time()
        eq(await lock2.acquire(), True)
        self.assertTrue(5 > lock2._loop.time() - start > 0.5)
        eq(await lock2.release(), True)

    def test_high_sleep_min(self):
        lock = self.client.lock('foo', blocking=1, sleep=2)
        self.assertEqual(lock.sleep, 1)

    async def test_releasing_lock_no_longer_owned_raises_error(self):
        key = self.randomkey()
        eq = self.assertEqual
        lock = self.client.lock(key)
        eq(await lock.acquire(), True)
        # manually change the token
        await self.client.set(key, 'a')
        await self.wait.assertRaises(LockError, lock.release)
        # even though we errored, the token is still cleared
        self.assertEqual(lock._token, None)

    async def test_context_manager(self):
        key = self.randomkey()
        lock2 = self.client.lock(key, blocking=False)
        async with self.client.lock(key):
            a = await lock2.acquire()
            self.assertFalse(a)
        a = await lock2.acquire()
        self.assertTrue(a)
        await lock2.release()

    async def test_context_manager_error(self):
        key = self.randomkey()
        lock2 = self.client.lock(key, blocking=False)

        async def _lock():
            async with lock2:
                pass

        async with self.client.lock(key):
            await self.wait.assertRaises(LockError, _lock)



import os
import sys
import glob
from importlib import import_module

import importlib.util

from .slugify import slugify


def _import_system_file(filename):
    module_name = slugify(filename[:-3], '.')
    spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def expand_star(mod_name):
    """Expand something like 'unuk.tasks.*' into a list of all the modules
    there.
    """
    expanded = []
    mod_dir = os.path.dirname(
        __import__(mod_name[:-2], {}, {}, ['']).__file__)
    for f in glob.glob1(mod_dir, "[!_]*.py"):
        expanded.append('%s.%s' % (mod_name[:-2], f[:-3]))
    return expanded


def import_modules(modules, safe=True):
    '''Safely import a list of *modules*
    '''
    all = []
    for mname in modules:
        if mname.endswith('.*'):
            to_load = expand_star(mname)
        else:
            to_load = [mname]
        for module in to_load:
            try:
                all.append(import_module(module))
            except ImportError:
                if not safe:
                    raise
    return all


def module_attribute(dotpath, default=None, safe=False):
    '''Load an attribute from a module.

    If the module or the attribute is not available, return the default
    argument if *safe* is `True`.
    '''
    if dotpath:
        bits = str(dotpath).split(':')
        try:
            if len(bits) == 2:
                attr = bits[1]
                module_name = bits[0]
            else:
                bits = bits[0].split('.')
                if len(bits) > 1:
                    attr = bits[-1]
                    module_name = '.'.join(bits[:-1])
                else:
                    raise ValueError('Could not find attribute in %s'
                                     % dotpath)

            module = import_module(module_name)
            return getattr(module, attr)
        except Exception:
            if not safe:
                raise
            return default
    else:
        if not safe:
            raise ImportError
        return default


py_extensions = set(('py', 'pyc', 'pyd', 'pyo'))


def py_file(name):
    bits = name.split('.')
    if len(bits) == 2 and bits[1] in py_extensions:
        return bits[0]
    elif len(bits) == 1:
        return name


def import_system_file(mod, safe=True):
    if os.path.isfile(mod):
        return _import_system_file(mod)
    else:
        try:
            return import_module(mod)
        except ImportError:
            mod2 = os.path.join(mod, '__init__.py')
            if os.path.isfile(mod2):
                return _import_system_file(mod2)
            elif not safe:
                raise
            pass



import time

from .string import *   # noqa

try:
    pypy = True
    import __pypy__     # noqa
except ImportError:     # pragma    nocover
    pypy = False


def identity(x):
    return x


default_timer = time.monotonic
as_iterator = identity



'''Tests the rpc middleware and utilities. It uses the calculator example.'''
import unittest
import asyncio

import pulsar
from pulsar import Config, get_actor
from pulsar.apps import MultiApp
from pulsar.apps.wsgi import WSGIServer

from tests.apps import dummy


class MultiWsgi(MultiApp):
    cfg = Config(bind=':0', rpc_bind=':0', bla='foo')

    def build(self):
        yield self.new_app(WSGIServer, callable=dummy)
        yield self.new_app(WSGIServer, 'rpc', callable=dummy)


class TestMultiApp(unittest.TestCase):

    def create(self, **params):
        # create the application
        return MultiWsgi(**params)

    def test_app(self):
        app = self.create(version='2.0')
        self.assertEqual(app.version, '2.0')
        self.assertTrue(app)
        apps = app.apps()
        self.assertEqual(len(apps), 2)
        self.assertEqual(apps[0].version, '2.0')
        self.assertEqual(apps[1].version, pulsar.__version__)

    def test_config(self):
        app = self.create(bind=':9999', unz='whatz')
        self.assertTrue(app)
        self.assertEqual(len(app.cfg.params), 4)
        app.apps()
        self.assertEqual(len(app.cfg.params), 2)
        self.assertEqual(app.cfg.bla, 'foo')
        self.assertEqual(app.cfg.unz, 'whatz')
        self.assertEqual(app.cfg.bind, ':9999')

    def test_timeout(self):
        app = self.create(timeout=22, name='pippo')
        self.assertEqual(app.name, 'pippo')
        self.assertEqual(app.cfg.timeout, 22)
        apps = app.apps()
        self.assertEqual(app.cfg.timeout, 22)
        self.assertNotEqual(app.cfg.rpc_timeout, 22)
        app1 = apps[0]
        self.assertEqual(app1.name, 'pippo')
        self.assertEqual(app1.cfg.timeout, 22)
        app2 = apps[1]
        self.assertEqual(app2.name, 'rpc_pippo')
        self.assertNotEqual(app2.cfg.timeout, 22)
        self.assertEqual(app2.cfg.timeout,
                         app2.cfg.settings['timeout'].default)

    def test_name(self):
        app = self.create()
        self.assertEqual(app.name, 'multiwsgi')
        app = self.create(name='bla')
        self.assertEqual(app.name, 'bla')
        apps = app.apps()
        self.assertEqual(apps[0].name, 'bla')
        self.assertEqual(apps[1].name, 'rpc_bla')

    async def test_start(self):
        arbiter = get_actor()
        app = self.create(name='pluto')
        self.assertTrue(app)
        self.assertFalse(arbiter.get_actor('pluto'))
        self.assertFalse(arbiter.get_actor('rpc_pluto'))
        # create the application
        await app.start()
        monitor1 = arbiter.get_actor('pluto')
        self.assertTrue(monitor1)
        self.assertTrue(monitor1.is_monitor())
        monitor2 = arbiter.get_actor('rpc_pluto')
        self.assertTrue(monitor2)
        self.assertTrue(monitor2.is_monitor())
        await asyncio.sleep(2)
        await monitor1.stop()
        await monitor2.stop()

    def test_config_copy(self):
        app = self.create()
        self.assertTrue(app)
        apps = app.apps()
        rpc = apps[1]
        cfg = rpc.cfg.copy()
        self.assertEqual(cfg.prefix, rpc.cfg.prefix)
        for name in cfg:
            self.assertTrue(name in rpc.cfg)



import unittest

from pulsar.apps.wsgi import Route


class Routes(unittest.TestCase):

    def testRoot(self):
        r = Route('/')
        self.assertFalse(r.is_leaf)
        self.assertEqual(r.rule, '')
        r = Route('////')
        self.assertFalse(r.is_leaf)
        self.assertEqual(r.rule, '')
        self.assertEqual(r.match(''), {})
        self.assertEqual(r.match('bee/'), {'__remaining__': 'bee/'})

    def testSimple(self):
        r = Route('bla/')
        self.assertFalse(r.is_leaf)
        self.assertEqual(r.level, 1)
        self.assertEqual(len(r.variables), 0)
        self.assertEqual(r.rule, 'bla/')
        self.assertEqual(r.match('bla/'), {})
        self.assertEqual(r.match('bladdd/'), None)
        self.assertEqual(r.match('bla/another/'),
                         {'__remaining__': 'another/'})

    def testSimpleLevel2(self):
        r = Route('bla/foo/')
        self.assertFalse(r.is_leaf)
        self.assertEqual(r.level, 2)
        self.assertEqual(len(r.variables), 0)
        self.assertEqual(r.rule, 'bla/foo/')
        self.assertEqual(r.match('bla/'), None)
        self.assertEqual(r.match('bla/foo/'), {})
        self.assertEqual(r.match('bla/fooddd/'), None)
        self.assertEqual(r.match('bla/foo/another/'),
                         {'__remaining__': 'another/'})

    def testMalformedRule1(self):
        self.assertRaises(ValueError, Route, '<bla')
        self.assertRaises(ValueError, Route, '<bla/')
        self.assertRaises(ValueError, Route, 'bla>/')
        self.assertRaises(ValueError, Route, '/<bla')

    def testMalformedRule2(self):
        self.assertRaises(ValueError, Route, '<foo>/<bla')
        self.assertRaises(ValueError, Route, 'ciao/<bla/')
        self.assertRaises(ValueError, Route, 'ahhh>/bla>/')
        self.assertRaises(ValueError, Route, 'ahhh/bla>/')

    def testMalformedRule3(self):
        self.assertRaises(ValueError, Route, '<foo>/<foo>')
        self.assertRaises(ValueError, Route, '<foo>/bla/<foo>')
        self.assertRaises(ValueError, Route, '<foo>/<bla>/<foo>')

    def testStringVariable(self):
        r = Route('<name>/')
        self.assertFalse(r.is_leaf)
        self.assertEqual(r.variables, set(['name']))
        self.assertEqual(r.breadcrumbs, ((True, 'name'),))
        self.assertEqual(r.rule, '<name>/')
        self.assertEqual(r.match('bla-foo/'), {'name': 'bla-foo'})
        self.assertEqual(r.match('bla/another/'),
                         {'name': 'bla', '__remaining__': 'another/'})
        self.assertEqual(r.url(name='luca'), '/luca/')

    def test2StringVariables(self):
        r = Route('<name>/<child>/')
        self.assertFalse(r.is_leaf)
        self.assertEqual(r.level, 2)
        self.assertEqual(r.variables, set(['name', 'child']))
        self.assertEqual(r.breadcrumbs, ((True, 'name'), (True, 'child')))
        self.assertEqual(r.rule, '<name>/<child>/')
        self.assertEqual(r.match('bla/foo/'), {'name': 'bla', 'child': 'foo'})
        self.assertEqual(r.match('bla/foo/another/'),
                         {'name': 'bla',
                          'child': 'foo',
                          '__remaining__': 'another/'})
        self.assertRaises(KeyError, r.url, name='luca')
        self.assertEqual(r.url(name='luca', child='joshua'), '/luca/joshua/')

    def test_add_dir_Leaf(self):
        r = Route('bla/')
        self.assertFalse(r.is_leaf)
        r2 = Route('foo')
        self.assertTrue(r2.is_leaf)
        r3 = r + r2
        self.assertEqual(r3.rule, 'bla/foo')
        self.assertTrue(r3.is_leaf)

    def testIntVariable(self):
        r = Route('<int:id>/')
        self.assertEqual(str(r), '/<int:id>/')
        self.assertEqual(r.variables, set(['id']))
        self.assertEqual(r.breadcrumbs, ((True, 'id'),))
        self.assertEqual(r.match('35/'), {'id': 35})
        self.assertEqual(r.url(id=1), '/1/')
        self.assertRaises(ValueError, r.url, id='bla')

    def testIntVariableFixDigits(self):
        r = Route('<int(2):id>/')
        self.assertEqual(str(r), '/<int(2):id>/')
        self.assertEqual(r.variables, set(['id']))
        self.assertEqual(r.breadcrumbs, ((True, 'id'),))
        self.assertEqual(r.match('35/'), {'id': 35})
        self.assertEqual(r.match('355/'), None)
        self.assertEqual(r.match('6/'), None)
        self.assertEqual(r.match('ch/'), None)
        self.assertEqual(r.url(id=13), '/13/')
        self.assertEqual(r.url(id=1), '/01/')
        self.assertRaises(ValueError, r.url, id=134)
        self.assertRaises(ValueError, r.url, id='bl')
        self.assertRaises(ValueError, r.url, id='bla')

    def testIntVariableMinMax(self):
        r = Route('<int(min=1):cid>/')
        self.assertEqual(str(r), '/<int(min=1):cid>/')
        self.assertEqual(r.variables, set(['cid']))
        self.assertEqual(r.breadcrumbs, ((True, 'cid'),))
        self.assertEqual(r.match('1/'), {'cid': 1})
        self.assertEqual(r.match('476876/'), {'cid': 476876})
        self.assertEqual(r.match('0/'), None)
        self.assertEqual(r.match('-5/'), None)
        self.assertEqual(r.url(cid=13), '/13/')
        self.assertEqual(r.url(cid=1), '/1/')
        self.assertRaises(ValueError, r.url, cid=0)
        self.assertRaises(ValueError, r.url, cid=-10)
        self.assertRaises(ValueError, r.url, cid='bla')

    def testPathVaiable(self):
        r = Route('bla/<path:rest>', defaults={'rest': ''})
        self.assertEqual(r.variables, set(['rest']))
        self.assertEqual(r.level, 2)
        self.assertTrue(r.is_leaf)
        self.assertEqual(r.match('bla/a/b/c.html'), {'rest': 'a/b/c.html'})
        self.assertEqual(r.match('bla/'), {'rest': ''})
        self.assertEqual(r.url(rest='a/'), '/bla/a/')
        self.assertEqual(r.url(), '/bla/')

    def testSplitRoot(self):
        r = Route('')
        self.assertEqual(r.level, 0)
        p, l = r.split()
        self.assertFalse(p.is_leaf)
        self.assertEqual(p.path, '/')
        self.assertEqual(l, None)
        r = Route('bla')
        p, l = r.split()
        self.assertFalse(p.is_leaf)
        self.assertTrue(l.is_leaf)
        self.assertEqual(p.path, '/')
        self.assertEqual(l.path, '/bla')
        r = Route('bla/')
        p, l = r.split()
        self.assertFalse(p.is_leaf)
        self.assertFalse(l.is_leaf)
        self.assertEqual(p.path, '/')
        self.assertEqual(l.path, '/bla/')

    def testSplitDir(self):
        r = Route('bla/foo/<id>/pluto/')
        self.assertEqual(r.level, 4)
        p, l = r.split()
        self.assertFalse(p.is_leaf)
        self.assertFalse(l.is_leaf)
        self.assertEqual(p.path, '/bla/foo/<id>/')
        self.assertEqual(l.path, '/pluto/')

    def testSplitLeaf(self):
        r = Route('bla/foo/<id>/pluto/leaf')
        p, l = r.split()
        self.assertFalse(p.is_leaf)
        self.assertTrue(l.is_leaf)
        self.assertEqual(p.path, '/bla/foo/<id>/pluto/')
        self.assertEqual(l.path, '/leaf')

    def testDefaults(self):
        r = Route('bla/<id>/add/<path:path>', {'path': ''})
        self.assertEqual(r.url(id=10), '/bla/10/add/')
        self.assertEqual(r.url(id=10, path='ciao/luca'),
                         '/bla/10/add/ciao/luca')

    def testDefaultsAddition(self):
        r = Route('add/<path:path>', {'path': ''})
        r2 = Route('bla/<id>/') + r
        self.assertEqual(r2.defaults, r.defaults)

    def test_add_string(self):
        r = Route('add/<path:path>', {'path': 'foo'})
        r2 = r + 'bla'
        self.assertEqual(r2.rule, 'add/<path:path>/bla')
        self.assertEqual(str(r2), '/add/<path:path>/bla')
        self.assertNotEqual(r2, '/add/<path:path>/bla')

    def test_empty_url(self):
        r = Route('')
        self.assertEqual(r.rule, '')
        self.assertEqual(r.url(), '/')
        self.assertEqual(r.path, '/')



import unittest

import pulsar


class TestApi(unittest.TestCase):

    def test_get_proxy(self):
        self.assertRaises(ValueError, pulsar.get_proxy, 'shcbjsbcjcdcd')
        self.assertEqual(pulsar.get_proxy('shcbjsbcjcdcd', safe=True), None)

    def test_bad_concurrency(self):
        # bla concurrency does not exists
        return self.wait.assertRaises(ValueError, pulsar.spawn, kind='bla')

    def test_actor_coverage(self):
        '''test case for coverage'''
        return self.wait.assertRaises(pulsar.CommandNotFound,
                                      pulsar.send, 'arbiter',
                                      'sjdcbhjscbhjdbjsj', 'bla')



import asyncio
from abc import ABC, abstractmethod

from ..utils.exceptions import LockError


class LockBase(ABC):
    """A asynchronous locking primitive associated to a given name.

    An asynchronous lock is in one of two states, 'locked' or 'unlocked'.
    It is created in the unlocked state. It has two basic methods,
    :meth:`.acquire` and :meth:`.release. When the state is unlocked,
    :meth:`.acquire` changes the state to locked and returns immediately.

    When the state is locked, :meth:`.acquire` wait
    until a call to :meth:`.release` changes it to unlocked,
    then the :meth:`.acquire` call resets it to locked and returns.

    .. attribute:: blocking

        The time to wait for the lock to be free when acquiring it.
        When False it does not block, when True it blocks forever,
        when a positive number blocks for ``blocking`` seconds.

    .. attribute:: timeout

        Free the lock after timeout seconds. If timeout is None (default)
        does not free the lock until ``release`` is called.
    """
    def __init__(self, name, *, loop=None, timeout=None, blocking=True):
        self.name = name
        self.timeout = timeout
        self.blocking = blocking
        self._loop = loop or asyncio.get_event_loop()

    @abstractmethod
    def locked(self):   # pragma    nocover
        """Return True if the lock is acquired
        """
        raise

    @abstractmethod
    async def acquire(self):    # pragma    nocover
        """Try to acquire the lock
        """
        raise

    @abstractmethod
    async def release(self):    # pragma    nocover
        """Release the lock
        """
        raise

    async def __aenter__(self):
        acquired = await self.acquire()
        if not acquired:
            raise LockError('Could not acquire lock')
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.locked():
            await self.release()


class Lock(LockBase):
    """An asynchronous lock
    """
    def __init__(self, name, *, loop=None, timeout=None, blocking=True):
        super().__init__(name, loop=loop, timeout=timeout, blocking=blocking)
        self._lock = _get_lock(self._loop, name)
        self._timeout_handler = None
        self._locked = False

    def locked(self):
        return self._locked

    async def acquire(self):
        try:
            timeout = self.blocking
            if timeout is True:
                timeout = None
            elif timeout is False:
                timeout = 0
            await asyncio.wait_for(self._lock.acquire(), timeout=timeout)
            self._schedule_timeout()
            self._locked = True
        except asyncio.TimeoutError:
            self._locked = False
        return self._locked

    async def release(self):
        self._cancel_lock()

    def _schedule_timeout(self):
        if self.timeout is not None:
            self._timeout_handler = self._loop.call_later(
                self.timeout, self._cancel_lock
            )

    def _cancel_lock(self):
        try:
            self._lock.release()
        except Exception:
            pass
        self._locked = False
        if self._timeout_handler:
            self._timeout_handler.cancel()
            self._timeout_handler = None


def _get_lock(loop, name):
    if not hasattr(loop, '_locks'):
        loop._locks = {}

    if name not in loop._locks:
        loop._locks[name] = asyncio.Lock(loop=loop)

    return loop._locks[name]



import unittest

from pulsar.apps import http


class TestBugs(unittest.TestCase):

    async def test_json(self):
        c = http.HttpClient()
        data = await c.get('https://api.bitfinex.com/v1/pubticker/BTCUSD')
        self.assertTrue(data)
